''''
to be added
'''

import copy
import re
from typing import Any, Callable, Dict, List, Optional, Tuple

from sastadev.alpino import getdehetwordinfo
from sastadev.basicreplacements import (basicexpansions, basicreplacements,
                                        getdisambiguationdict)
from sastadev.cleanCHILDEStokens import cleantokens
from sastadev.conf import settings
from sastadev.dedup import (cleanwordofnort, find_duplicates2,
                            find_janeenouduplicates, find_simpleduplicates,
                            find_substringduplicates2, getfilledpauses,
                            getprefixwords, getrepeatedtokens,
                            getunwantedtokens, nodesfindjaneenou)
from sastadev.deregularise import correctinflection
from sastadev.find_ngram import (Ngram, findmatches, ngram1, ngram2, ngram7,
                                 ngram10, ngram11, ngram16, ngram17)
from sastadev.iedims import getjeforms
from sastadev.lexicon import (WordInfo, de, dets, getwordinfo, het,
                              informlexicon, isa_namepart, known_word,
                              tswnouns)
from sastadev.macros import expandmacros
from sastadev.metadata import (Meta, bpl_indeze, bpl_node, bpl_none, bpl_word,
                               bpl_wordlemma, defaultbackplacement,
                               defaultpenalty, filled_pause, fstoken, intj,
                               janeenou, longrep, mkSASTAMeta, repeated,
                               repeatedjaneenou, repeatedseqtoken, shortrep,
                               substringrep, unknownsymbol)
from sastadev.sastatok import sasta_tokenize
from sastadev.sastatoken import Token, tokenlist2stringlist
from sastadev.sastatypes import (BackPlacement, MethodName, Nort, Penalty,
                                 Position, SynTree, UttId)
from sastadev.smallclauses import smallclauses
from sastadev.stringfunctions import (chatxxxcodes, consonants, deduplicate,
                                      endsinschwa, fullworddehyphenate,
                                      monosyllabic, vowels)
from sastadev.sva import getsvacorrections
from sastadev.tokenmd import TokenListMD, TokenMD, mdlist2listmd
from sastadev.treebankfunctions import (fatparse, find1, getattval,
                                        getnodeyield, showtree)

Correction = Tuple[List[Token], List[Meta]]
MetaCondition = Callable[[Meta], bool]

SASTA = 'SASTA'

hyphen = '-'
repetition = 'Repetition'

replacepattern = '{} [: {} ]'
metatemplate = '##META {} {} = {}'
slash = '/'
space = ' '

#: The constant *disambiguationdict* contains words that should be replaced by a
#: different word to avoid unwanted readings of the original word. It is filled by a
#: call to the function *getdisambiguationdict* from the module *basicreplacements*.
#:
#: .. autofunction:: basicreplacements::getdisambiguationdict
#:
disambiguationdict = getdisambiguationdict()

#: The constant *wrongdet_excluded_words* contains words that lead to incorrect
#: replacement of uter determiners (e.g. *die zijn* would be replaced by *dat zijn*) and
#: therefore have to be excluded from determiner replacement.
wrongdet_excluded_words = ['zijn', 'dicht', 'met', 'ik']


class Ngramcorrection:
    def __init__(self, ngram, fpositions, cpositions, metafunction):
        self.ngram: Ngram = ngram
        self.fpositions: Tuple[Position, Position] = fpositions
        self.cpositions: Tuple[Position, Position] = cpositions
        self.metafunction = metafunction


def mkmeta(att: str, val: str, type: str = 'text') -> str:
    result = metatemplate.format(type, att, val)
    return result


def anychars(chars: str) -> str:
    result = '[' + chars + ']'
    return result


def opt(pattern: str) -> str:
    result = '(' + pattern + ')?'
    return result


def replacement(inword: str, outword: str) -> str:
    result = replacepattern.format(inword, outword)
    return result


# duppattern = r'(.)\1{2,}'
# dupre = re.compile(duppattern)
#: The pattern *gaatiepattern* identifies words ending in *tie* preceded by at least a
#: vowel and optionally a consonant.
gaatiepattern = r'^.*' + anychars(vowels) + opt(anychars(consonants)) + 'tie$'
gaatiere = re.compile(gaatiepattern)
neutersgnoun = 'boekje'  # select here an unambiguous neuter noun


def skiptokens(tokenlist: List[Token], skiptokenlist: List[Token]) -> List[Token]:
    '''

    :param tokenlist:
    :param skiptokenlist:
    :return: a tokenlist identical to the input tokenlist but with the tokens that also occur with the same pos
    in skiptokenlist marked with skip=True
    '''
    skippositions = {token.pos for token in skiptokenlist}
    resultlist = []
    for token in tokenlist:
        if token.pos in skippositions:
            newtoken = Token(token.word, token.pos, skip=True)
        else:
            newtoken = token
        resultlist.append(newtoken)
    return resultlist


def ngramreduction(reducedtokens: List[Token], token2nodemap: Dict[Token, SynTree], allremovetokens: List[Token],
                   allremovepositions: List[Position], allmetadata: List[Meta], ngramcor: Ngramcorrection) \
        -> Tuple[List[Token], List[Token], List[Meta]]:
    # metadat function should still be added / abstracted
    (fb, fe) = ngramcor.fpositions
    (cb, ce) = ngramcor.cpositions
    reducedleaves = [token2nodemap[tok.pos] for tok in reducedtokens]

    vnwpvvnwpvmatches = findmatches(ngramcor.ngram, reducedleaves)
    allfalsestarttokens = []
    metadata = []
    for match in vnwpvvnwpvmatches:
        positions = [pos for pos in range(match[0], match[1])]
        falsestartpositions = [tok.pos for i, tok in enumerate(reducedtokens) if i in positions[fb:fe]]
        falsestarttokens = [tok for tok in reducedtokens if tok.pos in falsestartpositions]
        allfalsestarttokens += falsestarttokens
        correcttokenpositions = [tok.pos for i, tok in enumerate(reducedtokens) if i in positions[cb:ce]]
        correcttokens = [tok for tok in reducedtokens if tok.pos in correcttokenpositions]
        allremovetokens += falsestarttokens
        allremovepositions += falsestartpositions
        metadata += ngramcor.metafunction(falsestarttokens, falsestartpositions, correcttokens)
    reducedtokens = [tok for tok in reducedtokens if tok not in allfalsestarttokens]
    allmetadata += metadata
    return reducedtokens, allremovetokens, allmetadata


def reduce(tokens: List[Token], tree: Optional[SynTree]) -> Tuple[List[Token], List[Meta]]:
    if tree is None:
        settings.LOGGER.error('No tree for :{}\nNo reduction applied'.format(tokens))
        return ((tokens, []))

    tokennodes = tree.xpath('.//node[@pt or @pos]')
    tokennodesdict = {int(getattval(n, 'begin')): n for n in tokennodes}
    token2nodemap = {token.pos: tokennodesdict[token.pos] for token in tokens if keycheck(token.pos, tokennodesdict)}

    reducedtokens = tokens
    allmetadata = []

    allremovetokens = []
    allremovepositions = []

    # throw out unwanted symbols - -- # etc
    unwantedtokens = getunwantedtokens(reducedtokens)
    unwantedpositions = [tok.pos for tok in unwantedtokens]
    allremovetokens += unwantedtokens
    allremovepositions += unwantedpositions
    reducedtokens = [n for n in reducedtokens if n not in unwantedtokens]
    metadata = [mkSASTAMeta(token, token, 'ExtraGrammatical', unknownsymbol, 'Syntax') for token in unwantedtokens]
    allmetadata += metadata

    # remove  filled pauses

    filledpausetokens = getfilledpauses(reducedtokens)
    filledpausepositions = [token.pos for token in filledpausetokens]
    allremovetokens += filledpausetokens
    allremovepositions += filledpausepositions
    reducedtokens = [tok for tok in reducedtokens if tok not in filledpausetokens]
    reducednodes = [token2nodemap[tok.pos] for tok in reducedtokens if keycheck(tok.pos, token2nodemap)]
    metadata = [mkSASTAMeta(token, token, 'ExtraGrammatical', filled_pause, 'Syntax') for token in filledpausetokens]
    allmetadata += metadata

    # remove tsw incl goh och hé oke but not ja, nee, nou
    tswtokens = [n for n in reducedtokens if n.pos in token2nodemap
                 and getattval(token2nodemap[n.pos], 'pt') == 'tsw'
                 and getattval(token2nodemap[n.pos], 'lemma') not in {'ja', 'nee', 'nou'}
                 and getattval(token2nodemap[n.pos], 'lemma') not in tswnouns]
    tswpositions = [n.pos for n in tswtokens]
    allremovetokens += tswtokens
    allremovepositions == tswpositions
    reducedtokens = [n for n in reducedtokens if n not in tswtokens]
    metadata = [mkSASTAMeta(token, token, 'ExtraGrammatical', intj, 'Syntax') for token in tswtokens]
    allmetadata += metadata

    # find duplicatenode repetitions of ja, nee, nou
    janeenouduplicatenodes = find_janeenouduplicates(reducedtokens)
    allremovetokens += janeenouduplicatenodes
    reducedtokens = [n for n in reducedtokens if n not in janeenouduplicatenodes]
    reducednodes = [token2nodemap[tok.pos] for tok in reducedtokens if keycheck(tok.pos, token2nodemap)]
    metadata = [mkSASTAMeta(token, token, 'ExtraGrammatical', repeatedjaneenou, 'Syntax', subcat=repetition)
                for token in janeenouduplicatenodes]
    allmetadata += metadata

    # ASTA sec 6.3 p. 11
    # remove ja nee nou

    janeenounodes = nodesfindjaneenou(reducednodes)
    janeenoutokens = [tok for tok in reducedtokens if
                      keycheck(tok.pos, token2nodemap) and token2nodemap[tok.pos] in janeenounodes]
    janeenoupositions = [token.pos for token in janeenoutokens]
    allremovetokens += janeenoutokens
    allremovepositions += janeenoupositions
    reducedtokens = [tok for tok in reducedtokens if tok not in janeenoutokens]
    metadata = [mkSASTAMeta(token, token, 'ExtraGrammatical', janeenou, 'Syntax') for token in janeenoutokens]
    allmetadata += metadata

    # short repetitions
    def oldcond(x: Nort, y: Nort) -> bool:
        return len(cleanwordofnort(x)) / len(cleanwordofnort(y)) < .5 and not informlexicon(cleanwordofnort(x))

    def cond(x: Nort, y: Nort) -> bool:
        return len(cleanwordofnort(x)) / len(cleanwordofnort(
            y)) < .5  # check on lexicon put off actually two variants should be tried if the word is an existin gword

    shortprefixtokens = getprefixwords(reducedtokens, cond)
    shortprefixpositions = [token.pos for token in shortprefixtokens]
    repeatedtokens = getrepeatedtokens(reducedtokens, shortprefixtokens)
    allremovetokens += shortprefixtokens
    allremovepositions += shortprefixpositions
    metadata = [
        mkSASTAMeta(token, repeatedtokens[token], 'ExtraGrammatical', shortrep, 'Tokenisation', subcat=repetition) for
        token in reducedtokens if token in repeatedtokens]
    allmetadata += metadata
    reducedtokens = [tok for tok in reducedtokens if tok not in shortprefixtokens]

    # long repetitions
    def longcond(x: Nort, y: Nort) -> bool:
        return len(cleanwordofnort(x)) / len(cleanwordofnort(y)) >= .5 and not informlexicon(cleanwordofnort(x))

    longprefixtokens = getprefixwords(reducedtokens, longcond)
    longprefixpositions = [token.pos for token in longprefixtokens]
    repeatedtokens = getrepeatedtokens(reducedtokens, longprefixtokens)
    allremovetokens += longprefixtokens
    allremovepositions += longprefixpositions
    metadata = [
        mkSASTAMeta(token, repeatedtokens[token], 'ExtraGrammatical', longrep, 'Tokenisation', subcat=repetition) for
        token in reducedtokens if token in repeatedtokens]
    allmetadata += metadata
    reducedtokens = [tok for tok in reducedtokens if tok not in longprefixtokens]

    # find unknown words that are a substring of their successor
    substringtokens, _ = find_substringduplicates2(reducedtokens)
    substringpositions = [token.pos for token in substringtokens]
    repeatedtokens = getrepeatedtokens(reducedtokens, substringtokens)
    allremovetokens += substringtokens
    allremovepositions += substringpositions
    metadata = [mkSASTAMeta(token, repeatedtokens[token], 'ExtraGrammatical', substringrep, 'Tokenisation',
                            subcat=repetition) for token in reducedtokens if token in repeatedtokens]
    allmetadata += metadata
    reducedtokens = [tok for tok in reducedtokens if tok not in substringtokens]

    # simple duplicates
    dupnodetokens = find_simpleduplicates(reducedtokens)
    dupnodepositions = [token.pos for token in dupnodetokens]
    repeatedtokens = getrepeatedtokens(reducedtokens, dupnodetokens)
    allremovetokens += dupnodetokens
    allremovepositions += dupnodepositions
    metadata = [mkSASTAMeta(token, repeatedtokens[token], 'ExtraGrammatical',
                            repeated, 'Tokenisation', subcat=repetition) for token in reducedtokens if
                token in repeatedtokens]
    allmetadata += metadata
    reducedtokens = [tok for tok in reducedtokens if tok not in dupnodetokens]

    # duplicate sequences
    dupnodetokens, dupinfo = find_duplicates2(reducedtokens)
    dupnodepositions = [token.pos for token in dupnodetokens]
    duppairs = []
    for token in dupnodetokens:
        for othertok in reducedtokens:
            if token.pos in dupinfo.longdups and othertok.pos == dupinfo.longdups[token.pos]:
                nwt = othertok
                duppairs.append((token, nwt))
                break
    allremovetokens += dupnodetokens
    allremovepositions += dupnodepositions
    metadata = [mkSASTAMeta(token, nwt, 'ExtraGrammatical',
                            repeatedseqtoken, 'Tokenisation', subcat=repetition)
                for token, nwt in duppairs]
    allmetadata += metadata
    reducedtokens = [tok for tok in reducedtokens if tok not in dupnodetokens]

    # remove unknown words if open class DO NOT DO this
    # unknown_word_tokens = [tok for tok in reducedtokens if getattval(token2nodemap[tok.pos], 'pt') in openclasspts
    #                        and not (asta_recognised_wordnode(token2nodemap[tok.pos]))]
    # unknown_word_positions = [token.pos for token in unknown_word_tokens]
    # allremovetokens += unknown_word_tokens
    # allremovepositions += unknown_word_positions
    # metadata = [mkSASTAMeta(token, token, 'ExtraGrammatical',
    #                         unknownword, 'Tokenisation')
    #             for token in reducedtokens if token in unknown_word_tokens]
    # allmetadata += metadata
    # reducedtokens = [n for n in reducedtokens if n not in unknown_word_tokens]

    # ngram based cases

    # vnw pv vnw pv

    def metaf(falsestarttokens: List[Token], falsestartpositions: List[Position], correcttokens: List[Token]) \
            -> List[Meta]:
        return \
            [Meta('Retracing', 'Retracing with Correction', annotatedposlist=falsestartpositions,
                  annotatedwordlist=[c.word for c in falsestarttokens],
                  annotationposlist=[c.pos for c in correcttokens],
                  annotationwordlist=[c.word for c in correcttokens], cat='Retracing', subcat=None, source=SASTA,
                  penalty=defaultpenalty, backplacement=bpl_none)] + \
            [mkSASTAMeta(ftoken, ctoken, 'Retracing with Correction', fstoken, 'Retracing')
             for ftoken, ctoken in zip(falsestarttokens, correcttokens)]

    vnwpvvnwpvcor = Ngramcorrection(ngram1, (0, 2), (2, 4), metaf)
    reducedtokens, allremovetokens, allmetadata = ngramreduction(reducedtokens, token2nodemap, allremovetokens,
                                                                 allremovepositions, allmetadata, vnwpvvnwpvcor)

    vzdetvzdetcor = Ngramcorrection(ngram2, (0, 2), (2, 4), metaf)
    reducedtokens, allremovetokens, allmetadata = ngramreduction(reducedtokens, token2nodemap, allremovetokens,
                                                                 allremovepositions, allmetadata, vzdetvzdetcor)

    vgdetvgdetcor = Ngramcorrection(ngram7, (0, 2), (2, 4), metaf)
    reducedtokens, allremovetokens, allmetadata = ngramreduction(reducedtokens, token2nodemap, allremovetokens,
                                                                 allremovepositions, allmetadata, vgdetvgdetcor)
    vnwipvjxpvjvnwi = Ngramcorrection(ngram10, (0, 2), (3, 5), metaf)
    reducedtokens, allremovetokens, allmetadata = ngramreduction(reducedtokens, token2nodemap, allremovetokens,
                                                                 allremovepositions, allmetadata, vnwipvjxpvjvnwi)
    lemilemjlemilemj = Ngramcorrection(ngram11, (0, 2), (3, 5), metaf)
    reducedtokens, allremovetokens, allmetadata = ngramreduction(reducedtokens, token2nodemap, allremovetokens,
                                                                 allremovepositions, allmetadata, lemilemjlemilemj)

    dinjdknj = Ngramcorrection(ngram16, (0, 2), (3, 5), metaf)
    reducedtokens, allremovetokens, allmetadata = ngramreduction(reducedtokens, token2nodemap, allremovetokens,
                                                                 allremovepositions, allmetadata, dinjdknj)

    tevtev = Ngramcorrection(ngram17, (0, 2), (2, 4), metaf)
    reducedtokens, allremovetokens, allmetadata = ngramreduction(reducedtokens, token2nodemap, allremovetokens,
                                                                 allremovepositions, allmetadata, tevtev)

    # reducedleaves = [token2nodemap[tok.pos] for tok in reducedtokens]
    #
    # vnwpvvnwpvmatches = findmatches(ngram1, reducedleaves)
    # allfalsestarttokens = []
    # metadata = []
    # for match in vnwpvvnwpvmatches:
    #     positions = [pos for pos in range(match[0], match[1])]
    #     falsestartpositions = [tok.pos for i, tok in enumerate(reducedtokens) if i in positions[0:2]]
    #     falsestarttokens = [tok for tok in reducedtokens if tok.pos in falsestartpositions]
    #     allfalsestarttokens += falsestarttokens
    #     correcttokenpositions = [tok.pos for i, tok in enumerate(reducedtokens) if i in positions[2:5]]
    #     correcttokens = [tok for tok in reducedtokens if tok.pos in correcttokenpositions]
    #     allremovetokens += falsestarttokens
    #     allremovepositions += falsestartpositions
    #     metadata += [Meta('Retracing', 'Retracing with Correction',  annotatedposlist=falsestartpositions,
    #                  annotatedwordlist=[c.word for c in falsestarttokens], annotationposlist=[c.pos for c in correcttokens],
    #                  annotationwordlist=[c.word for c in correcttokens], cat='Retracing', subcat=None, source=SASTA,
    #                     penalty=defaultpenalty, backplacement=bpl_none)]
    #     metadata += [mkSASTAMeta(ftoken, ctoken, 'Retracing with Correction', fstoken,  'Retracing', )
    #                 for ftoken, ctoken  in zip(falsestarttokens, correcttokens)]
    #
    # reducedtokens = [tok for tok in reducedtokens if tok not in allfalsestarttokens]
    # allmetadata += metadata

    skipmarkedtokens = skiptokens(tokens, allremovetokens)

    # return (reducedtokens, allremovetokens, allmetadata)
    return (skipmarkedtokens, allmetadata)


def keycheck(key: Any, dict: Dict[Any, Any]) -> bool:
    if key not in dict:
        settings.LOGGER.error('key {}  not in dictionary. Contents of dictionary:'.format(key))
        for akey, val in dict.items():
            valbgn = getattval(val, 'begin')
            valpt = getattval(val, 'pt')
            valword = getattval(val, 'word')
            valstr = '{}:{}:{}'.format(valbgn, valpt, valword)
            settings.LOGGER.error('{}={}'.format(akey, valstr))
    return key in dict


def combinesorted(toklist1: List[Token], toklist2: List[Token]) -> List[Token]:
    result = toklist1 + toklist2
    sortedresult = sorted(result, key=lambda tok: tok.pos)
    return sortedresult


# def getcorrection(utt, tree=None, interactive=False):
#     # NOT used anymore!!!!
#
#     allmetadata = []
#     rawtokens = sasta_tokenize(utt)
#     wordlist = tokenlist2stringlist(rawtokens)
#
#     tokens, metadata = cleantokens(rawtokens, repkeep=False)
#     allmetadata += metadata
#     tokensmd = TokenListMD(tokens, [])
#
#     # reducedtokens, allremovedtokens, metadata = reduce(tokens)
#     # allremovedtokens, metadata = reduce(tokens)
#     skipmarkedtokens, metadata = reduce(tokens, tree)
#     # reducedtokensmd = TokenListMD(reducedtokens, [])
#     reducedtokensmd = TokenListMD(skipmarkedtokens, [])
#
#     alternativemds = getalternatives(reducedtokensmd, tree, 0)
#     # alternativemds = getalternatives(tokensmd, allremovedtokens, tree, 0)
#     # unreducedalternativesmd = [TokenListMD(combinesorted(alternativemd.tokens, allremovedtokens), alternativemd.metadata) for alternativemd in alternativemds]
#
#     # correctiontokensmd = unreducedalternativesmd[-1] if unreducedalternativesmd != [] else tokensmd
#     correctiontokensmd = alternativemds[-1] if alternativemds != [] else tokensmd
#
#     correction = tokenlist2stringlist(correctiontokensmd.tokens)
#     allmetadata += correctiontokensmd.metadata
#
#     result = (correction, allmetadata)
#     return result


def getcorrections(rawtokens: List[Token], method: MethodName, tree: Optional[SynTree] = None,
                   interactive: bool = False) -> List[Correction]:
    allmetadata = []
    # rawtokens = sasta_tokenize(utt)
    wordlist = tokenlist2stringlist(rawtokens)
    utt = space.join(wordlist)
    origutt = utt
    #print(utt)

    # check whether the tree has the same yield
    origtree = tree
    treeyield = getnodeyield(tree)
    treewordlist = [getattval(n, 'word') for n in treeyield]

    if treewordlist != wordlist:
        revisedutt = space.join(wordlist)
        tree = fatparse(revisedutt, rawtokens)

    tokens, metadata = cleantokens(rawtokens, repkeep=False)
    allmetadata += metadata
    tokensmd = TokenListMD(tokens, [])

    # reducedtokens, allremovedtokens, metadata = reduce(tokens)
    reducedtokens, metadata = reduce(tokens, tree)
    reducedtokensmd = TokenListMD(reducedtokens, [])
    allmetadata += metadata

    # alternativemds = getalternatives(reducedtokensmd, tree, 0)
    alternativemds = getalternatives(reducedtokensmd, method, tree, '0')
    # unreducedalternativesmd = [TokenListMD(combinesorted(alternativemd.tokens, allremovedtokens), alternativemd.metadata) for alternativemd in alternativemds]

    intermediateresults = alternativemds if alternativemds != [] else [tokensmd]

    results = []
    for ctmd in intermediateresults:
        # correction = tokenlist2stringlist(ctmd.tokens)
        correction = ctmd.tokens
        themetadata = allmetadata + ctmd.metadata
        results.append((correction, themetadata))
    return results


# def getalternatives(origtokensmd, method, llremovedtokens, tree, uttid):
def getalternatives(origtokensmd: TokenListMD, method: MethodName, tree: SynTree, uttid: UttId):
    tokensmd = explanationasreplacement(origtokensmd, tree)
    if tokensmd is None:
        tokensmd = origtokensmd

    tokens = tokensmd.tokens
    allmetadata = tokensmd.metadata
    #newtokens = []
    #alternatives = []
    alternativetokenmds = {}
    validalternativetokenmds = {}
    tokenctr = 0
    for token in tokens:
        tokenmd = TokenMD(token, allmetadata)
        alternativetokenmds[tokenctr] = getalternativetokenmds(tokenmd, method, tokens, tokenctr, tree, uttid)
        validalternativetokenmds[tokenctr] = getvalidalternativetokenmds(tokenmd, alternativetokenmds[tokenctr])
        tokenctr += 1

    # get all the new token sequences
    tokenctr = 0
    lvalidalternativetokenmds = len(validalternativetokenmds)
    altutts: List[List[TokenMD]] = [[]]
    newutts = []
    while tokenctr < lvalidalternativetokenmds:
        for tokenmd in validalternativetokenmds[tokenctr]:
            for utt in altutts:
                newutt = copy.copy(utt)
                newutt.append(tokenmd)
                newutts.append(newutt)
        altutts = newutts
        newutts = []
        tokenctr += 1

    # now turn each sequence of (token, md) pairs into a pair (tokenlist, mergedmetadata)
    newaltuttmds = []
    for altuttmd in altutts:
        if altuttmd != []:
            newaltuttmd = mdlist2listmd(altuttmd)
            newaltuttmds.append(newaltuttmd)

    # basic expansions
    allalternativemds = newaltuttmds  # put off, taken care of in getvalidalternatives:  + [tokensmd]

    newresults = []
    for uttmd in allalternativemds:
        expansionmds = getexpansions(uttmd)
        newresults += expansionmds
    allalternativemds += newresults

    # combinations of tokens or their alternatives: de kopje, de stukkie, heeft gevalt

    newresults = []
    for uttmd in allalternativemds:
        # utterance = space.join([token.word for token in uttmd.tokens])
        utterance, _ = mkuttwithskips(uttmd.tokens)
        fatntree = fatparse(utterance, uttmd.tokens)
        newresults += getwrongdetalternatives(uttmd, fatntree, uttid)
    allalternativemds += newresults

    newresults = []
    for uttmd in allalternativemds:
        # utterance = space.join([token.word for token in uttmd.tokens])
        utterance, _ = mkuttwithskips(uttmd.tokens)
        # reducedtokens = [t for t in uttmd.tokens if not t.skip]
        # reduceduttmd = TokenListMD(reducedtokens, uttmd.metadata)
        fatntree = fatparse(utterance, uttmd.tokens)
        debug = False
        if debug:
            showtree(fatntree)
        uttalternativemds = getsvacorrections(uttmd, fatntree, uttid)
        newresults += uttalternativemds
    allalternativemds += newresults

    newresults = []
    for uttmd in allalternativemds:
        # utterance = space.join([token.word for token in uttmd.tokens])
        utterance, _ = mkuttwithskips(uttmd.tokens)
        fatntree = fatparse(utterance, uttmd.tokens)
        newresults += correctPdit(uttmd, fatntree, uttid)
    allalternativemds += newresults

    newresults = []
    for uttmd in allalternativemds:
        utterance, _ = mkuttwithskips(uttmd.tokens)
        fatntree = fatparse(utterance, uttmd.tokens)
        newresults += smallclauses(uttmd, fatntree)
        # showtree(fatntree, text='fatntree')
    allalternativemds += newresults

    # final check whether the alternatives are improvements. It is not assumed that the original tokens is included in the alternatives
    finalalternativemds = lexcheck(tokensmd, allalternativemds)

    return finalalternativemds


skiptemplate = "[ @skip {} ]"


def oldmkuttwithskips(tokens: List[Token], toskip: List[Token]) -> str:
    sortedtokens = sorted(tokens, key=lambda x: x.pos)
    resultlist = []
    for token in sortedtokens:
        if token in toskip:
            resultlist.append(skiptemplate.format(token.word))
        else:
            resultlist.append(token.word)
    result = space.join(resultlist)
    return result


def mkuttwithskips(tokens: List[Token], delete: bool = True) -> Tuple[str, List[Position]]:
    sortedtokens = sorted(tokens, key=lambda x: x.pos)
    resultlist = []
    tokenposlist = []
    for token in sortedtokens:
        if token.skip:
            if not delete:
                resultlist.append(skiptemplate.format(token.word))
                tokenposlist.append(token.pos)
        else:
            resultlist.append(token.word)
            tokenposlist.append(token.pos)
    result = space.join(resultlist)

    return result, tokenposlist


def OLDgetexpansions(uttmd: TokenListMD) -> List[TokenListMD]:
    '''

    :param uttmd: the list of tokens in the utterance with its metadata
    :return: zero or more alternative lists of tokens with metadata

    The function *getexpansions* generates alternative tokenlists plus metadata for
    words that are contractions and must be expanded into a sequence of multiple tokens.
    It checks whether a word is a contraction by checking whether it occurs in the
    dictionary *basicexpansions* from the module *basicreplacements*

        .. autodata:: basicreplacements::basicexpansions
            :no-value:

    '''

    expansionfound = False
    newtokens = []
    tokenctr = 0
    #newtokenctr = 0
    tokenposlist = []
    newmd = uttmd.metadata
    for tokenctr, token in enumerate(uttmd.tokens):
        if token.word.lower() in basicexpansions:
            expansionfound = True
            for (rlist, c, n, v) in basicexpansions[token.word.lower()]:
                rlisttokenctr = 0
                for rlisttokenctr, rw in enumerate(rlist):
                    if rlisttokenctr == 0:
                        newtoken = Token(rw, token.pos)
                    else:
                        newtoken = Token(rw, token.pos, subpos=rlisttokenctr)
                    newtokens.append(newtoken)
                    tokenposlist.append(token.pos)
                    nwt = Token(space.join(rlist), token.pos)
                meta1 = mkSASTAMeta(token, nwt, n, v, c, subcat=None, penalty=defaultpenalty,
                                    backplacement=bpl_none)
                newmd.append(meta1)

        else:
            newtoken = Token(token.word, token.pos)
            newtokens.append(newtoken)
            tokenposlist.append(token.pos)

    # adapt the metadata
    if expansionfound:
        meta2 = Meta('OrigCleanTokenPosList', tokenposlist, annotatedposlist=[],
                     annotatedwordlist=[], annotationposlist=tokenposlist,
                     annotationwordlist=[], cat='Tokenisation', subcat=None, source=SASTA, penalty=defaultpenalty,
                     backplacement=bpl_none)
        newmd.append(meta2)
        result = [TokenListMD(newtokens, newmd)]
    else:
        result = []

    return result


def getsingleitemexpansions(token: Token, intokenposlist) -> List[Tuple[TokenListMD, List[int]]]:
    lcword = token.word.lower()
    outtokenposlist = copy.copy(intokenposlist)
    if lcword in basicexpansions:
        results = []
        for (rlist, c, n, v, p) in basicexpansions[lcword]:
            outtokenposlist = copy.copy(intokenposlist)
            newtokens = []
            newmd = []
            for rlisttokenctr, rw in enumerate(rlist):

                if rlisttokenctr == 0:
                    newtoken = Token(rw, token.pos)
                else:
                    newtoken = Token(rw, token.pos, subpos=rlisttokenctr)
                newtokens.append(newtoken)
                outtokenposlist.append(token.pos)
                nwt = Token(space.join(rlist), token.pos)
            meta1 = mkSASTAMeta(token, nwt, n, v, c, subcat=None, penalty=p,
                                backplacement=bpl_none)
            newmd.append(meta1)
            result = (TokenListMD(newtokens, newmd), outtokenposlist)
            results.append(result)
    else:
        outtokenposlist.append(token.pos)
        results = [(TokenListMD([token], []), outtokenposlist)]

    return results


def combine(headresult: Tuple[TokenListMD, List[int]], tailresult: Tuple[TokenListMD, List[int]]) \
        -> Tuple[TokenListMD, List[int]]:
    '''

    :param headresult: an  expansion result for the head
    :param tailresult: an expansion result for the tail
    :return: the combination of the head result and the tailresult

    The function *combine* combines a result for the head of an input token list with a result of the tail of
    the input token list.
    It simply concatenates the token lists of the results, and the metadata of the results,
    and generates the tokenposlist for their combination.

    '''
    newtokens = headresult[0].tokens + tailresult[0].tokens
    newmd = headresult[0].metadata + tailresult[0].metadata
    newtokenposlist = tailresult[1]
    result = (TokenListMD(newtokens, newmd), newtokenposlist)
    return result


def getexpansions2(tokenlist: List[Token], intokenposlist: List[int]) -> List[Tuple[TokenListMD, List[int]]]:
    '''

    :param tokenlist: the list of tokens in the utterance
    :param intokenposlist: the list of token positions so far, initially the empty list
    :return: zero or more alternative lists of tuples of
       * tokens with metadata
       * accumulated list of token positions

    The function *getexpansions2* generates alternative tokenlists plus metadata for
    words that are contractions and must be expanded into a sequence of multiple tokens.
    It applies the function *getsingleitemexpansions* to the head (first) element of the *tokenlist* and
    recursively applies itself to the tail of *intokenlist*, after which it combines the results by the *combine* function.

        .. autofunction::  corrector::getsingleitemexpansions

        .. autofunction:: corrector::combine

    It checks whether a word is a contraction by checking whether it occurs in the
    dictionary *basicexpansions* from the module *basicreplacements*

        .. autodata:: basicreplacements::basicexpansions
            :no-value:

    '''
    finalresults = []
    if tokenlist == []:
        outtokenposlist = copy.copy(intokenposlist)
        finalresults = [(TokenListMD([], []), outtokenposlist)]
    else:
        headresults = getsingleitemexpansions(tokenlist[0], intokenposlist)
        for headresult in headresults:
            tailresults = getexpansions2(tokenlist[1:], headresult[1])
            results = [combine(headresult, tailresult) for tailresult in tailresults]
            finalresults += [(TokenListMD(result[0].tokens, result[0].metadata), result[1]) for result in results]
    return finalresults


def gettokenyield(tokens: List[Token]) -> str:
    words = [token.word for token in tokens]
    result = space.join(words)
    return result


def getexpansions(uttmd: TokenListMD) -> List[TokenListMD]:
    '''

    :param uttmd: the list of tokens in the utterance with its metadata
    :return: a possibly empty list of alternative lists of tokens with metadata

    The function *getexpansions* generates alternative tokenlists plus metadata for
    words that are contractions and must be expanded into a sequence of multiple tokens.

    It does so by a call to the function *getexpansions2*, which recursively generates all alternatives with expansions:

    .. autofunction:: corrector::getexpansions2

    '''
    newtokenmds = []

    results = getexpansions2(uttmd.tokens, [])
    for result in results:
        result0yield = gettokenyield(result[0].tokens)
        uttmdyield = gettokenyield(uttmd.tokens)
        if result0yield != uttmdyield:  # otherwise we get unnecessary and undesired duplicates
            tokenposlist = result[1]
            meta2 = Meta('OrigCleanTokenPosList', tokenposlist, annotatedposlist=[],
                         annotatedwordlist=[], annotationposlist=tokenposlist,
                         annotationwordlist=[], cat='Tokenisation', subcat=None, source=SASTA, penalty=0,
                         backplacement=bpl_none)
            newmd = result[0].metadata
            newmd.append(meta2)
            newtokenmd = TokenListMD(result[0].tokens, newmd)
            newtokenmds.append(newtokenmd)

    return newtokenmds

    # adapt the metadata
    # finalresults = []
    # for result in results:
    #     meta2 = Meta('OrigCleanTokenPosList', tokenposlist, annotatedposlist=[],
    #                  annotatedwordlist=[], annotationposlist=tokenposlist,
    #                  annotationwordlist=[], cat='Tokenisation', subcat=None, source=SASTA, penalty=defaultpenalty,
    #                  backplacement=bpl_none)
    #     newmd = result.metadata
    #     newmd.append(meta2)
    #     finalresult = [TokenListMD(result.tokens, newmd)]
    #     finalresults.append(finalresult)


def lexcheck(intokensmd: TokenListMD, allalternativemds: List[TokenListMD]) -> List[TokenListMD]:
    finalalternativemds = [intokensmd]
    for alternativemd in allalternativemds:
        diff_found = False
        include = True
        intokens = intokensmd.tokens
        outtokens = alternativemd.tokens
        if len(intokens) != len(outtokens):
            diff_found = True
        else:
            for (intoken, outtoken) in zip(intokens, outtokens):
                if intoken != outtoken:
                    diff_found = True
                    if not known_word(outtoken.word):
                        include = False
                        break
        if diff_found and include:
            finalalternativemds.append(alternativemd)
    return finalalternativemds


# moved to metadata
# def mkSASTAMeta(token, nwt, name, value, cat, subcat=None, penalty=defaultpenalty, backplacement=defaultbackplacement):
#    result = Meta(name, value, annotatedposlist=[token.pos],
#                     annotatedwordlist=[token.word], annotationposlist=[nwt.pos],
#                     annotationwordlist=[nwt.word], cat=cat, subcat=subcat, source=SASTA, penalty=penalty,
#                     backplacement=backplacement)
#    return result


def updatenewtokenmds(newtokenmds: List[TokenMD], token: Token, newwords: List[str], beginmetadata: List[Meta],
                      name: str, value: str, cat: str, subcat: Optional[str] = None,
                      penalty: Penalty = defaultpenalty, backplacement: BackPlacement = defaultbackplacement) \
        -> List[TokenMD]:
    for nw in newwords:
        nwt = Token(nw, token.pos)
        meta = mkSASTAMeta(token, nwt, name=name, value=value, cat=cat, subcat=subcat, penalty=penalty,
                           backplacement=backplacement)
        metadata = [meta] + beginmetadata
        newwordtokenmd = TokenMD(nwt, metadata)
        newtokenmds.append(newwordtokenmd)
    return newtokenmds


def gettokensplusxmeta(tree: SynTree) -> Tuple[List[Token], List[Meta]]:
    '''
    converts the origutt into  list of xmeta elements
    :param tree: input tree
    :return: list of xmeta elements
    '''
    origutt = find1(tree, './/meta[@name="origutt"]/@value')
    tokens1 = sasta_tokenize(origutt)
    tokens2, metadata = cleantokens(tokens1, repkeep=False)
    return tokens2, metadata


def findxmetaatt(xmetalist: List[Meta], name: str, cond: MetaCondition = lambda x: True) -> Optional[Meta]:
    cands = [xm for xm in xmetalist if xm.name == name and cond(xm)]
    if cands == []:
        result = None
    else:
        result = cands[0]
    return result


def tokenreplace(oldtokens: List[Token], newtoken: Token) -> List[Token]:
    newtokens = []
    for token in oldtokens:
        if token.pos == newtoken.pos:
            newtokens.append(newtoken)
        else:
            newtokens.append(token)
    return newtokens


def explanationasreplacement(tokensmd: TokenListMD, tree: SynTree) -> Optional[TokenListMD]:
    # interpret single word explanation as replacement # this will work only after retokenistion of the origutt
    result = None
    origmetadata = tokensmd.metadata
    xtokens, xmetalist = gettokensplusxmeta(tree)
    explanations = [xm for xm in xmetalist if xm.name == 'Explanation']
    newtokens = copy.deepcopy(xtokens)
    newmetadata = origmetadata + xmetalist
    for explanation in explanations:
        newwordlist = explanation.annotationwordlist
        oldwordlist = explanation.annotatedwordlist
        tokenposlist = explanation.annotatedposlist
        if len(newwordlist) == 1 and len(tokenposlist) == 1 and len(oldwordlist) == 1:
            newword = newwordlist[0]
            oldwordpos = tokenposlist[0]
            oldword = oldwordlist[0]
            newtoken = Token(newword, oldwordpos)
            oldtoken = Token(oldword, oldwordpos)
            if known_word(newword):
                newtokens = tokenreplace(newtokens, newtoken)
                bpl = bpl_node if known_word(oldword) else bpl_word
                meta = mkSASTAMeta(oldtoken, newtoken, name='ExplanationasReplacement',
                                   value='ExplanationasReplacement',
                                   cat='Lexical Error', backplacement=bpl)
                newmetadata.append(meta)
                result = TokenListMD(newtokens, newmetadata)
    return result


# some words are known but very unlikely as such
#: The constant *specialdevoicingwords* contains known words that start with a
#: voiceless consonant for which the word starting with the corresponding voiced
#: consonant is much more likely in young children's speech.
specialdevoicingwords = {'fan'}


def initdevoicing(token: Token, voiceless: str, voiced: str, newtokenmds: List[TokenMD], beginmetadata: List[Meta]) \
        -> List[TokenMD]:
    '''
    The function *initdevoicing* takes as input *token*, checks whether it is an
    unknown word or a special known word. If the token's word starts with *voiceless*
    it creates a newword with the tokens's word initial character replaced by *voiced*.
    If the result is a known word, *newtokenmds* is updated with the new replacement
    and *beginmetadata*, and it returns *newtokenmds*.

    A known word is *special* if it is contained in the variable *specialdevoicingwords*.

    .. autodata:: corrector::specialdevoicingwords

    '''
    # initial s -> z, f -> v
    if not known_word(token.word.lower()) or token.word.lower() in specialdevoicingwords:
        if token.word[0] == voiceless:
            newword = voiced + token.word[1:]
            if known_word(newword):
                newtokenmds = updatenewtokenmds(newtokenmds, token, [newword], beginmetadata,
                                                name='Pronunciation Variant',
                                                value='Initial {} devoicing'.format(voiced),
                                                cat='Pronunciation', backplacement=bpl_word)

    return newtokenmds


def getalternativetokenmds(tokenmd: TokenMD, method: MethodName, tokens: List[Token], tokenctr: int,
                           tree: SynTree, uttid: UttId) -> List[TokenMD]:
    token = tokenmd.token
    beginmetadata = tokenmd.metadata
    newtokenmds: List[TokenMD] = []

    if token.skip:
        return newtokenmds

    # decapitalize initial token  except when it is a known name
    if tokenctr == 0 and token.word.istitle() and not isa_namepart(token.word):
        newword = token.word.lower()

        newtokenmds = updatenewtokenmds(newtokenmds, token, [newword], beginmetadata,
                                        name='Character Case', value='Lower case', cat='Orthography')

    # dehyphenate
    if not known_word(token.word) and hyphen in token.word:
        newwords = fullworddehyphenate(token.word, known_word)
        newtokenmds = updatenewtokenmds(newtokenmds, token, newwords, beginmetadata,
                                        name='Dehyphenation', value='Dehyphenation', cat='Pronunciation',
                                        backplacement=bpl_word)

    # deduplicate jaaaaa -> ja; heeeeeel -> heel
    if not known_word(token.word):
        newwords = deduplicate(token.word, known_word, exceptions=chatxxxcodes)
        newtokenmds = updatenewtokenmds(newtokenmds, token, newwords, beginmetadata,
                                        name='Emphasis', value='Phoneme lengthening', cat='Pronunciation',
                                        backplacement=bpl_word)

    # basic replacements replace as by als, isse by is
    # here come the replacements
    if token.word.lower() in basicreplacements:
        for (r, c, n, v, p) in basicreplacements[token.word.lower()]:
            newwords = [r]
            newtokenmds = updatenewtokenmds(newtokenmds, token, newwords, beginmetadata,
                                            name=n, value=v, cat=c, backplacement=bpl_word, penalty=p)
    # next replaced by conditions in basicreplacements.py
    # if token.word.lower() in {'wel', 'niet'}:
    #     beginval = str(token.pos)
    #     wordnode = find1(tree, f'.//node[@pt="ww" and @begin="{beginval}"]')
    #     if wordnode is not None:
    #         newwords = ['ietsjes']
    #         newtokenmds = updatenewtokenmds(newtokenmds, token, newwords, beginmetadata,
    #                                         name='Disambiguation', value='Avoid unknown reading',
    #                                         cat='Lexicon', backplacement=bpl_wordlemma)

    moemoetxpath = './/node[@lemma="moe" and @pt!="n" and not(%onlywordinutt%) and (@rel="--" or @rel="dp" or @rel="predm" or @rel="nucl")]'
    expanded_moemoetxpath = expandmacros(moemoetxpath)
    if token.word.lower() == 'moe' and tree.xpath(expanded_moemoetxpath) != []:
        newwords = ['moet']
        newtokenmds = updatenewtokenmds(newtokenmds, token, newwords, beginmetadata,
                                        name='Informal pronunciation', value='Final t-deletion', cat='Pronunciation',
                                        backplacement=bpl_word)

    # dee -> deze of deed
    if token.word.lower() == 'dee':
        newwords = ['deze']
        newtokenmds = updatenewtokenmds(newtokenmds, token, newwords, beginmetadata,
                                        name='Wrong pronunciation', value='Coda reduction', cat='Pronunciation',
                                        backplacement=bpl_word, penalty=5)
        newwords = ['deed']
        newtokenmds = updatenewtokenmds(newtokenmds, token, newwords, beginmetadata,
                                        name='Informal pronunciation', value='Final t-deletion', cat='Pronunciation',
                                        backplacement=bpl_word)

    # find document specific replacements

    # find organisation specific replacements

    # find childes replacements, preferably with vocabulary from the same age

    # gaatie
    if not known_word(token.word):
        newwords = gaatie(token.word)
        newtokenmds = updatenewtokenmds(newtokenmds, token, newwords, beginmetadata,
                                        name='Word combination', value='Cliticisation', cat='Pronunciation',
                                        backplacement=bpl_none)

    # extend to gaat-ie

    # dediacritisize

    # iedims
    if token.word.endswith('ie') or token.word.endswith('ies'):
        newwords = getjeforms(token.word)
        newtokenmds = updatenewtokenmds(newtokenmds, token, newwords, beginmetadata,
                                        name='RegionalForm', value='ieDim', cat='Morphology', backplacement=bpl_word)

    # overregularised verb forms: gevalt -> gevallen including  incl  wrong verb forms: gekeekt -> gekeken
    if not known_word(token.word):
        nwms = correctinflection(token.word)
        for nw, metavalue in nwms:
            newtokenmds += updatenewtokenmds(newtokenmds, token, [nw], beginmetadata,
                                             name='InflectionError', value=metavalue, cat='Morphology',
                                             backplacement=bpl_word)

    # wrong verb forms: gekeekt -> gekeken: done!

    # me ze (grote/oudere/ kleine) moeder /vader/zusje/ broer -> mijn me s done by Alpino, here we do ze
    # next xpath does not work because it must be preceded by a . !!
    # zexpathmodel = """//node[@word="ze" and @begin={begin} and (@rel="--"  or (@rel="obj1" and parent::node[@cat="pp"])) and @end = ancestor::node[@cat="top"]/descendant::node[@pt="n"]/@begin]"""
    if token.word.lower() == 'ze' or token.word.lower() == 'su':
        # find the node that corresponds to this token in the tree
        # zexpath = zexpathmodel.format(begin=str(tokenctr))
        # zenode = find1(tree, zexpath)
        tokennodes = getnodeyield(tree)
        zenode = tokennodes[tokenctr]
        nexttoken = tokens[
            tokenctr + 1]  # do not take it from the tree because it may have been replaced by something else, e.g. avoid: ze dee -> ze deed -/-> z'n deed!
        zerel = getattval(zenode, 'rel')
        zeparent = zenode.getparent()
        zeparentcat = getattval(zeparent, 'cat')
        # nextpt = getattval(nextnode, 'pt')
        nexttokeninfo = getwordinfo(nexttoken.word.lower())
        nexttokenpts = {pt for (pt, _, _, _) in nexttokeninfo}
        if (zerel == '--' or zerel == 'mwp' or (zerel == 'obj1' and zeparentcat == 'pp')) and 'n' in nexttokenpts:
            newword = "z'n"
            newtokenmds = updatenewtokenmds(newtokenmds, token, [newword], beginmetadata,
                                            name='Pronunciation Variant', value='N-less informal possessive pronoun',
                                            cat='Pronunciation', backplacement=bpl_word)

    # e-> e(n)
    enexceptions = {'inne', 'mette', 'omme', 'oppe', 'vanne'}
    if not known_word(
            token.word) and token.word.lower() not in basicreplacements and token.word.lower() not in enexceptions:
        if endsinschwa(token.word) and not monosyllabic(token.word):
            newword = token.word + 'n'
            if known_word(newword):
                newtokenmds = updatenewtokenmds(newtokenmds, token, [newword], beginmetadata,
                                                name='Pronunciation Variant', value='N-drop after schwa',
                                                cat='Pronunciation', backplacement=bpl_word)

    # initial s -> z
    newtokenmds = initdevoicing(token, 's', 'z', newtokenmds, beginmetadata)
    # initial f -> v
    newtokenmds = initdevoicing(token, 'f', 'v', newtokenmds, beginmetadata)

    # replaceambiguous words with one reading not known by the child by a nonambiguous word with the same properties
    if method in {'tarsp', 'stap'}:
        if token.word.lower() in disambiguationdict:
            cond, newword = disambiguationdict[token.word.lower()]
            if cond(token, tree):
                newtokenmds = updatenewtokenmds(newtokenmds, token, [newword], beginmetadata,
                                                name='Disambiguation', value='Avoid unknown reading',
                                                cat='Lexicon', backplacement=bpl_wordlemma)

    # ...en -> e: groten  -> grote (if adjective); goten -> grote

    # drop e at the end incl duplicated consonants (ooke -> ook; isse -> is ? DOne, basicreplacements

    # losse e -> een / het / de

    for newtokenmd in newtokenmds:
        morenewtokenmds = getalternativetokenmds(newtokenmd, method, tokens, tokenctr, tree, uttid)
        newtokenmds += morenewtokenmds

    return newtokenmds


def getvalidalternativetokenmds(tokenmd: TokenMD, newtokenmds: List[TokenMD]) -> List[TokenMD]:
    validnewtokenmds = [tokenmd for tokenmd in newtokenmds if known_word(tokenmd.token.word)]
    # and now we add the original tokenmd
    validnewtokenmds += [tokenmd]
    return validnewtokenmds


def gaatie(word: str) -> List[str]:
    '''
    The function *gaatie* replaces  a word that matches with  *gaatiepattern*  (e.g.
    *gaatie*) by a sequence of two words where the first word equals word[:-2] (
    *gaat*) and is a known word and the second word equals word[-2:] (*ie*).

    .. autodata:: corrector::gaatiepattern
    '''
    results = []
    if gaatiere.match(word):
        if informlexicon(word[:-2]):  # and if it is a verb this is essential because now tie is also split into t ie
            result = space.join([word[:-2], word[-2:]])
            results.append(result)
    return results


def getwrongdetalternatives(tokensmd: TokenListMD, tree: SynTree, uttid: UttId) -> List[TokenListMD]:
    '''
    The function *getwrongdetalternatives* takes as input a TokenListMD *tokensmd*,  the
    original parse of the utterance (*tree*) and the *uttid* of the utterance.

    It inspects each token in the token list of *tokensmd* that should not be skipped
    and that is a utrum determiner. If the token that immediately follows this
    determiner is not a token to be ignored we obtain the gender properties of the
    token's word (there can be multiple if it is ambiguous). If one of the properties
    is neuter gender and none is uter, then the uter determiner is replaced by its neuter
    variant as a new alternative.

    The token following must be ignored if it has the property *skip=True* or if it
    belongs to words that would lead to wrong corrections, as specified in the constant
    *wrongdet_excluded_words*:

    .. autodata:: corrector::wrongdet_excluded_words

    The properties of the token following are determined by the function
    *getdehetwordinfo* from the module *alpino*:

    .. autofunction:: alpino::getdehetwordinfo
    '''
    correctiondone = False
    tokens = tokensmd.tokens
    metadata = tokensmd.metadata
    ltokens = len(tokens)
    tokenctr = 0
    newtokens = []
    while tokenctr < ltokens:
        token = tokens[tokenctr]
        if not token.skip and token.word in dets[de] and tokenctr < ltokens - 1:
            nexttoken = tokens[tokenctr + 1]
            # we want to exclude some words
            if nexttoken.skip:
                wordinfos: List[WordInfo] = []
            elif nexttoken.word in wrongdet_excluded_words:
                wordinfos = []
            else:
                wordinfos, _ = getdehetwordinfo(nexttoken.word)
            if wordinfos != []:
                for wordinfo in wordinfos:  # if there are multiple alternatives we overwrite and therefore get the last alternative
                    (pos, dehet, infl, lemma) = wordinfo
                    if dehet == het and infl in ['e', 'de']:
                        # newcurtoken = replacement(token, swapdehet(token))
                        newcurtokenword = swapdehet(token.word)
                        newcurtoken = Token(newcurtokenword, token.pos)
                        meta = mkSASTAMeta(token, newcurtoken, name='GrammarError', value='deheterror', cat='Error',
                                           backplacement=bpl_node)
                        metadata.append(meta)
                        correctiondone = True
                    else:
                        newcurtokenword = token.word
                newtokens.append(Token(newcurtokenword, token.pos))
            else:
                newcurtokenword = token.word
                newtokens.append(token)
        else:
            newtokens.append(token)
        tokenctr += 1
    result = TokenListMD(newtokens, metadata)
    if correctiondone:
        results = [result]
    else:
        results = []
    return results


def getindezemwp(prevtokennode: SynTree, tokennode: SynTree) -> bool:
    ok = True
    ok = ok and getattval(prevtokennode, 'lemma') in {'in'}
    ok = ok and getattval(prevtokennode, 'rel') in {'mwp'}
    ok = ok and getattval(tokennode, 'lemma') in {'deze'}
    ok = ok and getattval(tokennode, 'rel') in {'mwp'}
    return ok


def correctPdit(tokensmd: TokenListMD, tree: SynTree, uttid: UttId) -> List[TokenListMD]:
    '''
    The function *correctPdit* replaces demonstrative pronouns immediately preceded by
    an adposition by the pronoun *hem*. It sets the value of the *backplacement*
    attribute of the metadata to *bpl_node* so that it will be replaced again by the
    original node after the parse has been done, unless the original parse contained the multiword
    unit *in deze*. Then the *backplacement* attribute gets the value *bpl_indeze* so
    that in a later stage some special replacements will be performed.

    The function takes as input:

    * *tokensmd* of type *TpkenListMD* : the list of tokens wit hassociated metadata

    *tree* of type *SynTree*: the parse of the original utterance

    *uttid* of type *UttId*: the utterance identifier of the utterance (currently not
    used)

    It yields a list  containing the alternatives generated (of type List[TokenListMD].
    '''
    correctiondone = False
    tokennodes = getnodeyield(tree)
    tokens = tokensmd.tokens
    metadata = tokensmd.metadata
    newtokens = []
    tokenctr = 0
    nonskiptokenctr = 0
    prevtoken = None
    for token in tokens:
        tokennode = next(filter(lambda x: getattval(x, 'begin') == str(token.pos + token.subpos), tokennodes), None)
        tokenlemma = getattval(tokennode, 'lemma')
        if not token.skip and prevtoken is not None and not prevtoken.skip and tokenlemma in {'dit', 'dat', 'deze',
                                                                                              'die'}:
            tokenrel = getattval(tokennode, 'rel')
            tokenpt = getattval(tokennode, 'pt')
            prevtokennode = tokennodes[nonskiptokenctr - 1] if tokenctr > 0 else None
            if prevtokennode is not None:
                prevpt = getattval(prevtokennode, 'pt')
                prevparent = prevtokennode.getparent()
                prevparentrel, prevparentcat = getattval(prevparent, 'rel'), getattval(prevparent, 'cat')
                indezemwp = getindezemwp(prevtokennode, tokennode)
                if (prevpt == 'vz' and prevparentcat != 'pp' and tokenrel not in {'det'} and tokenpt == 'vnw') or \
                        indezemwp:
                    newtoken = Token('hem', token.pos, subpos=token.subpos)
                    bpl = bpl_indeze if indezemwp else bpl_node
                    meta = mkSASTAMeta(token, newtoken, name='parsed as', value='hem', cat='AlpinoImprovement',
                                       backplacement=bpl, penalty=15)
                    metadata.append(meta)
                    newtokens.append(newtoken)
                    correctiondone = True
                else:
                    newtokens.append(token)
            else:
                newtokens.append(token)
        else:
            newtokens.append(token)
        tokenctr += 1
        if not token.skip:
            nonskiptokenctr += 1
        prevtoken = token
    result = TokenListMD(newtokens, metadata)
    if correctiondone:
        results = [result]
    else:
        results = []
    return results


def parseas(w: str, code: str) -> str:
    result = '[ @add_lex {} {} ]'.format(code, w)
    return result


def swapdehet(dedet: str) -> Optional[str]:
    if dedet in dets[de]:
        deindex = dets[de].index(dedet)
    else:
        deindex = -1
    if dedet in dets[het]:
        hetindex = dets[het].index(dedet)
    else:
        hetindex = -1
    if deindex >= 0:
        result = dets[het][deindex]
    elif hetindex >= 0:
        result = dets[de][hetindex]
    else:
        result = None
    return result


def outputalternatives(tokens, alternatives, outfile):
    for el in alternatives:
        print(tokens[el], slash.join(alternatives[el]), file=outfile)


def mkchatutt(intokens: List[str], outtokens: List[str]) -> List[str]:
    result = []
    for (intoken, outtoken) in zip(intokens, outtokens):
        newtoken = intoken if intoken == outtoken else replacement(intoken, outtoken)
        result.append(newtoken)
    return result


def altmkchatutt(intokens: List[str], outtoken: str) -> List[str]:
    result = []
    for intoken in intokens:
        newtoken = intoken if intoken == outtoken else replacement(intoken, outtoken)
        result.append(newtoken)
    return result



from collections import Counter
from copy import deepcopy
from typing import List, Optional, Tuple

from sastadev.allresults import AllResults
from sastadev.lexicon import getwordinfo, getwordposinfo
from sastadev.sastatypes import QId, SynTree, UttId
from sastadev.stringfunctions import realwordstring
from sastadev.treebankfunctions import getattval, getnodeyield, getuttid

lpad = 3
zero = '0'
astamaxwordcount = 300

excluded_lemmas = ['gevallen', 'gewinnen']

nounqid = 'A021'
lexqid = 'A018'
samplesizeqid = 'A045'
mluxqid = 'A029'
pvqid = 'A024'
delpvqid = 'A009'
subpvqid = 'A032'
kqid = 'A013'
mqid = 'A020'
tijdfoutpvqid = 'A041'
nounlemmaqid = 'A046'
verblemmaqid = 'A049'

specialform = 'Special Form'
errormarking = 'Error Marking'

#: The varibale (constant) *mdnamemdxpathtemplate* is an Xpath template to find
#: metadata (xmeta) with  name=*mdname* and value=*mdvalue*
mdnamemdxpathtemplate = """.//xmeta[@name="{mdname}" and @value="{mdvalue}"]"""

ptposxpathtemplate = './/node[@pt and @begin="{position}"]'


def mdbasedquery(stree: SynTree, mdname: str, mdvalue: str) -> List[SynTree]:
    '''
    The function *mdbasedquery* searches for metadata in *stree* with name = *mdname*
    and value = *mdvalue*. It then obtains the position of the node to which the
    metadata apply, and next finds all nodes with that position as value for its *begin* attribute.
    '''
    mdnamemdxpath = mdnamemdxpathtemplate.format(mdname=mdname, mdvalue=mdvalue)
    mdnamemds = stree.xpath(mdnamemdxpath)
    results = []
    for mdnamemd in mdnamemds:
        annotatedposstr = mdnamemd.attrib['annotatedposlist']
        if annotatedposstr != '':
            mdbeginval = annotatedposstr[1:-1]
            ptposxpath = ptposxpathtemplate.format(position=mdbeginval)
            newresults = stree.xpath(ptposxpath)
            results += newresults

    return results


def neologisme(stree: SynTree) -> List[SynTree]:
    '''
    The function *neologisme* identifies the nodes for which the CHAT error marking "[*
    n]" or the special form "@n" applies. It uses the function *mdbasedquery* to
    achieve this.

    .. autofunction:: ASTApostfunctions::mdbasedquery

    '''
    results1 = mdbasedquery(stree, errormarking, "['n']")
    results2 = mdbasedquery(stree, specialform, '@n')
    results = results1 + results2
    return results


def sempar(stree: SynTree) -> List[SynTree]:
    '''
    The function *sempar* identifies the nodes for which the CHAT error marking "[*
    s]"  applies. It uses the function *mdbasedquery* to achieve this.

    .. autofunction:: ASTApostfunctions::mdbasedquery


    '''
    results = mdbasedquery(stree, errormarking, "['s']")
    return results


def phonpar(stree: SynTree) -> List[SynTree]:
    '''
    The function *phonpar* identifies the nodes for which the CHAT error marking "[*
    p]"  applies. It uses the function *mdbasedquery* to achieve this.

    .. autofunction:: ASTApostfunctions::mdbasedquery

    '''
    results = mdbasedquery(stree, errormarking, "['p']")
    return results


def sumctr(ctr):
    result = sum(ctr.values())
    return result


def wordcountperutt(allresults):
    lemmas = getalllemmas(allresults)
    wordcounts = {uttid: sum(ctr.values()) for uttid, ctr in lemmas.items()}
    ignorewordcounts = deepcopy(
        allresults.coreresults[
            samplesizeqid]) if samplesizeqid in allresults.coreresults else Counter()  # samplesize
    ignorewordcounts += allresults.coreresults[
        mluxqid] if mluxqid in allresults.coreresults else Counter()  # mlux
    # ignorewordcounts += allresults.coreresults['A050'] if 'A050' in allresults.coreresults else Counter() # echolalie covered by mlux
    result = {}
    for uttid in wordcounts:
        tosubtract = ignorewordcounts[uttid] if uttid in ignorewordcounts else 0
        result[uttid] = wordcounts[uttid] - tosubtract
    # remove uttids which have 0 words
    result = {uttid: ctr for uttid, ctr in result.items() if ctr != 0}
    return result


def finietheidsindex(allresults, _):
    allpvs = allresults.coreresults[
        pvqid] if pvqid in allresults.coreresults else Counter()
    subpvs = allresults.coreresults[
        subpvqid] if subpvqid in allresults.coreresults else Counter()
    delpvs = allresults.coreresults[
        delpvqid] if delpvqid in allresults.coreresults else Counter()
    tijdfoutpvs = allresults.coreresults[
        tijdfoutpvqid] if tijdfoutpvqid in allresults.coreresults else Counter()
    foutepvs = subpvs + delpvs + tijdfoutpvs
    allpvcount = sumctr(allpvs)
    foutepvcount = sumctr(foutepvs)
    okpvcount = allpvcount - foutepvcount
    if allpvcount == 0:
        result = 0
    else:
        result = okpvcount / allpvcount
    return result


def countwordsandcutoff(allresults, _):
    # @@to be adapted
    result = (None, 0)
    if 'A047' in allresults.postresults:
        paddedlist = []
        for key, val in allresults.postresults['A047'].items():
            paddedkey = key.rjust(lpad, zero)
            paddedlist.append((paddedkey, val))
        sortedlist = sorted(paddedlist)
        wc = 0
        for key, val in sortedlist:
            if wc + val > astamaxwordcount:
                result = (key, wc)
                break
            else:
                wc += val
                result = (None, wc)
    return result


def KMcount(allresults, _):
    Kcount = sumctr(allresults.coreresults[kqid]) if kqid in allresults.coreresults else 0
    Mcount = sumctr(allresults.coreresults[mqid]) if mqid in allresults.coreresults else 0
    result = Kcount + Mcount
    return result


def old_old_getlemmas(allresults, _):
    allmatches = allresults.allmatches
    allresults.postresults['A046'] = Counter()
    for el in allmatches:
        (qid, uttid) = el
        if qid in ['A021', 'A018']:
            for amatch in allmatches[el]:
                # theword = normalizedword(amatch[0])
                theword = getattval(amatch[0], 'lemma')
                allresults.postresults['A046'].update([(theword, uttid)])
    return allresults


def getlemmas(allresults, _):
    result = getcondlemmas(allresults, _, lambda qid: qid in [nounqid, lexqid])
    return result


def getnounlemmas(allresults, _):
    '''
    The function *getnounlemmas* uses the function *getposlemmas* applied to
    *allresults* and the query identifier for nouns to obtain the lemmas
    for nouns.

    .. autofunction:: ASTApostfunctions::getposlemmas

    '''
    result = getposlemmas(allresults, nounqid)
    return result


def getlexlemmas(allresults, _):
    '''
    The function *getlexlemmas* uses the function *getposlemmas* applied to
    *allresults* and the query identifier for lexical verbs to obtain the lemmas
    for lexical verbs.

    .. autofunction:: ASTApostfunctions::getposlemmas
    '''
    result = getposlemmas(allresults, lexqid)
    return result


def realword(node):
    result = getattval(node, 'pt') not in ['let']
    return result


def getalllemmas(allresults):
    result = {}
    if allresults.annotationinput:
        for uttid in allresults.allutts:
            lemmas = [bgetlemma(w) for w in allresults.allutts[uttid] if
                      realwordstring(w)]
            result[uttid] = Counter(lemmas)
    else:
        for syntree in allresults.analysedtrees:
            uttid = getuttid(syntree)
            lemmas = [getattval(node, 'lemma') for node in getnodeyield(syntree) if
                      realword(node)]
            result[uttid] = Counter(lemmas)
    return result


def old_getlemmas(allresults, _):
    allmatches = allresults.allmatches
    result = Counter()
    for el in allmatches:
        (qid, uttid) = el
        if qid in ['A021', 'A018']:
            for amatch in allmatches[el]:
                # theword = normalizedword(amatch[0])
                theword = getattval(amatch[0], 'lemma')
                result.update([(theword, uttid)])
    return result


def oldgetcondlemmas(allresults, _, cond):
    allmatches = allresults.allmatches
    result = Counter()
    for el in allmatches:
        (qid, uttid) = el
        if cond(qid):
            for amatch in allmatches[el]:
                # theword = normalizedword(amatch[0])
                theword = getattval(amatch[0], 'lemma')
                result.update([(theword, uttid)])
    return result


# not used anymore, contains an error
def getcondlemmas(allresults, _, cond):
    result = Counter()
    if allresults.annotationinput:
        for qid in allresults.exactresults:
            if cond(qid):
                for (uttid, position) in allresults.exactresults[qid]:
                    word = allresults.allutts[uttid][position - 1]
                    if qid == 'A021':
                        pos = 'n'
                    elif qid == 'A018':
                        pos = 'ww'
                    else:
                        pos = None
                    lemma = bgetlemma(word, pos)
                    result.update([(lemma, qid, uttid)])

    else:
        allmatches = allresults.allmatches
        for el in allmatches:
            (qid, uttid) = el
            if cond(qid):
                for amatch in allmatches[el]:
                    # theword = normalizedword(amatch[0])
                    theword = getattval(amatch[0], 'lemma')
                    result.update([(theword, uttid)])
    return result


def getposfromqid(qid):
    if qid == 'A021':
        pos = 'n'
    elif qid == 'A018':
        pos = 'ww'
    else:
        pos = None
    return pos


def getposlemmas(allresults: AllResults, posqid: QId) -> List[Tuple[str, UttId]]:
    '''
    The function *getposlemmas* obtains the lemmas from *allresults* that have been
    found by a query with identifier *posqid*.

    The lemma is obtained from the parse tree if there is one, otherwise (in case the
    input was an annotation form) from the lexicon (CELEX).
    '''
    result = Counter()
    if allresults.annotationinput:
        for (uttid, position) in allresults.exactresults[posqid]:
            word = allresults.allutts[uttid][position - 1]
            pos = getposfromqid(posqid)
            lemma = bgetlemma(word, pos)
            result.update([(lemma, uttid)])
    else:
        allmatches = allresults.allmatches
        for el in allmatches:
            (qid, uttid) = el
            if qid == posqid:
                for amatch in allmatches[el]:
                    # theword = normalizedword(amatch[0])
                    theword = getattval(amatch[0], 'lemma')
                    result.update([(theword, uttid)])
    return result


def bgetlemma(word: str, pos: Optional[str] = None):
    if pos is None:
        wordinfos = getwordinfo(word)
        if wordinfos == []:
            lemma = word
        else:
            filteredwordinfos = [wi for wi in wordinfos if wi[3] not in excluded_lemmas]
            if filteredwordinfos == []:
                lemma = wordinfos[0][3]
            else:
                lemma = filteredwordinfos[0][3]
    else:
        wordinfos = getwordposinfo(word, pos)
        if wordinfos == []:
            lemma = word
        else:
            filteredwordinfos = [wi for wi in wordinfos if wi[3] not in excluded_lemmas]
            if filteredwordinfos == []:
                lemma = wordinfos[0][3]
            else:
                lemma = filteredwordinfos[0][3]
    return lemma




from typing import Optional

#import find1, iswordnode, getattval
import sastadev.stringfunctions as strf
import sastadev.treebankfunctions as tbf
from sastadev.sastatypes import SynTree

#import CHAT_Annotation as schat  # put off because it causes an error: AttributeError: module 'CHAT_Annotation' has no attribute 'wordpat'


space = ' '
CHAT_explanation = 'Explanation'
explannwordlistxpath = f'.//xmeta[@name="{CHAT_explanation}"]/@annotationwordlist'
explannposlistxpath = f'.//xmeta[@name="{CHAT_explanation}"]/@annotationposlist'


def finalmultiwordexplanation(stree: SynTree) -> Optional[str]:
    #get the multiword explanation and the last tokenposition is occupies

    explannwrdliststr = tbf.find1(stree, explannwordlistxpath)
    explannwrdlist = strf.string2list(explannwrdliststr, quoteignore=True)

    explannposliststr = tbf.find1(stree, explannposlistxpath)
    explannposlist = strf.string2list(explannposliststr)

    ismultiword = len(explannwrdlist) > 1

    if ismultiword:
        # any token in the tree with begin > last tokenposition of explanation can only be an interpunction sign
        # check whether it is the last one ignoring interpunction

        explannposlast = int(explannposlist[-1])

        explisfinal = True
        for node in stree.iter():
            if explisfinal:
                if tbf.iswordnode(node):
                    beginstr = tbf.getattval(node, 'begin')
                    if beginstr != '':
                        begin = int(beginstr)
                        if begin > explannposlast:
                            nodept = tbf.getattval(node, 'pt')
                            if nodept not in {'let'}:
                                explisfinal = False

        if explisfinal:
            result = space.join(explannwrdlist)
        else:
            result = None
        return result



'''
This module offers the function get_annotations() to obtain a dictionary with the annotations from a file
for the moment at the utteranceid level, to be extended to the wordposition per uttid level

and the function read_annotations() to obtain a score dictionary with queryid as keys and Counter() as values
'''

# todo
# -additional columns unaligned treatment and generalisation
# -code alternatives and replacemtne extensions
# =codes written without spaces?

import re
from collections import Counter, defaultdict
from typing import Any, Dict, List, Match, Optional, Pattern, Tuple

from sastadev import xlsx
from sastadev.conf import settings
from sastadev.readmethod import itemseppattern
from sastadev.sastatypes import (FileName, Item, Level, Position, QId,
                                 QueryDict, UttId, UttWordDict)

varitem = ''

txtext = ".txt"
comma = ","
space = ' '
tsvext = '.tsv'
commaspace = ', '
tab = '\t'
all_levels = set()
literallevels = ['lemma']

semicolon = ';'
labelsep = semicolon

wordcolheaderpattern = r'^\s*word\d+\s*$'
wordcolheaderre = re.compile(wordcolheaderpattern)
firstwordcolheaderpattern = r'^\s*word0*1\s*$'
firstwordcolheaderre = re.compile(firstwordcolheaderpattern)

speakerheaders = ['speaker', 'spreker', 'spk']
uttidheaders = ['id', 'utt', 'uttid']
levelheaders = ['level']
stagesheaders = ['fases', 'stages']
commentsheaders = ['comments', 'commentaar']
unalignedheaders = ['unaligned']


def nested_dict(n: int, type: type):  # I do not know how to characterize the result type Dict n times deep endin gwith values of type type
    if n == 1:
        return defaultdict(type)
    else:
        return defaultdict(lambda: nested_dict(n - 1, type))


def clean(label: str) -> str:
    result = label
    result = result.lstrip()
    result = result.rstrip()
    result = result.lower()
    return result


def getlabels(labelstr: str, patterns: Tuple[Pattern, Pattern]) -> List[str]:
    results = []
    (pattern, fullpattern) = patterns
    if fullpattern.match(labelstr):
        ms = pattern.finditer(labelstr)
        results = [m.group(0) for m in ms if m.group(0) not in ' ;,-/']
    else:
        results = []
        ms = pattern.finditer(labelstr)
        logstr = str([m.group(0) for m in ms if m.group(0) not in ' ;,-'])
        # print('Cannot interpret {};  found items: {}'.format(labelstr,logstr), file=sys.stderr)
        settings.LOGGER.warning('Cannot interpret %s; found items: %s', labelstr, logstr)
        # exit(-1)
    return results


def iswordcolumn(str: str) -> Optional[Match[str]]:
    result = wordcolheaderre.match(str.lower())
    return result


def isfirstwordcolumn(str: str) -> Optional[Match[str]]:
    result = firstwordcolheaderre.match(str.lower())
    return result


def enrich(labelstr: str, lcprefix: str) -> str:
    labels = labelstr.split(labelsep)
    newlabels = []
    for label in labels:
        cleanlabel = clean(label)
        if label != "" and lcprefix != "":
            newlabels.append(lcprefix + ":" + cleanlabel)
        else:
            newlabels.append(cleanlabel)
    result = labelsep.join(newlabels)
    return result


def getcleanlevelsandlabels(thelabelstr: str, thelevel: str, prefix: str, patterns: Tuple[Pattern, Pattern]) \
        -> List[Tuple[str, str]]:
    results: List[Tuple[str, str]] = []
    lcthelabelstr = thelabelstr.lower()
    lcprefix = prefix.lower().strip()
    lcthelabelstr = enrich(lcthelabelstr, lcprefix)
    thelabels = getlabels(lcthelabelstr, patterns)
    for thelabel in thelabels:
        if thelabel != "":
            cleanlabel = thelabel
            cleanlevel = clean(thelevel)
            result = (cleanlevel, cleanlabel)
            results.append(result)

    return results


# def oldget_annotations(infilename, patterns):
#     '''
#     Reads the file with name filename in SASTA Annotation Format
#     :param infilename:
#     :param patterns
#     :return: a dictionary  with as  key a tuple (level, item) and as value a Counter  with key uttid and value its count
#     '''
#
#     thedata = defaultdict(list)
#     exactdata = defaultdict(list)
#     cdata = {}
#
#     # To open Workbook
#     wb = xlrd.open_workbook(infilename)
#     sheet = wb.sheet_by_index(0)
#
#     startrow = 0
#     startcol = 0
#     headerrow = 0
#     headers = {}
#     lastrow = sheet.nrows
#     lastcol = sheet.ncols
#     #    firstwordcol = 2
#     #    lastwordcol = lastcol - 4
#     levelcol = 1
#     uttidcol = 0
#     stagescol = -1
#     commentscol = -1
#
#     uttlevel = 'utt'
#
#     for rowctr in range(startrow, lastrow):
#         if rowctr == headerrow:
#             for colctr in range(startcol, lastcol):
#                 headers[colctr] = sheet.cell_value(rowctr, colctr)
#                 if iswordcolumn(headers[colctr]):
#                     lastwordcol = colctr
#                     if isfirstwordcolumn(headers[colctr]):
#                         firstwordcol = colctr
#                 elif clean(headers[colctr]) in speakerheaders:
#                     spkcol = colctr
#                 elif clean(headers[colctr]) in uttidheaders:
#                     uttidcol = colctr
#                 elif clean(headers[colctr]) in levelheaders:
#                     levelcol = colctr
#                 elif clean(headers[colctr]) in stagesheaders:
#                     stagescol = colctr
#                 elif clean(headers[colctr]) in commentsheaders:
#                     commentscol = colctr
#         else:
#             if sheet.cell_value(rowctr, uttidcol) != "":
#                 uttid = str(int(sheet.cell_value(rowctr, uttidcol)))
#             thelevel = sheet.cell_value(rowctr, levelcol)
#             thelevel = clean(thelevel)
#             all_levels.add(thelevel)
#             for colctr in range(firstwordcol, sheet.ncols):
#                 if thelevel in literallevels and colctr != stagescol and colctr != commentscol:
#                     thelabel = sheet.cell_value(rowctr, colctr)
#                     if colctr > lastwordcol:
#                         tokenposition = 0
#                     else:
#                         tokenposition = colctr - firstwordcol + 1
#                     thedata[(thelevel, thelabel)].append(uttid)
#                     exactdata[(cleanlevel, cleanlabel)].append((uttid, tokenposition))
#                 elif thelevel != uttlevel and colctr != stagescol and colctr != commentscol:
#                     thelabelstr = sheet.cell_value(rowctr, colctr)
#                     thelevel = sheet.cell_value(rowctr, levelcol)
#                     if lastwordcol + 1 <= colctr < sheet.ncols:
#                         # prefix = headers[colctr] aangepast om het simpeler te houden
#                         prefix = ""
#                     else:
#                         prefix = ""
#                     cleanlevelsandlabels = getcleanlevelsandlabels(thelabelstr, thelevel, prefix, patterns)
#                     if colctr > lastwordcol:
#                         tokenposition = 0
#                     else:
#                         tokenposition = colctr - firstwordcol + 1
#                     for (cleanlevel, cleanlabel) in cleanlevelsandlabels:
#                         thedata[(cleanlevel, cleanlabel)].append(uttid)
#                         exactdata[(cleanlevel, cleanlabel)].append((uttid, tokenposition))
#     # wb.close() there is no way to close the workbook
#     for atuple in thedata:
#         cdata[atuple] = Counter(thedata[atuple])
#     return cdata
#
#
# def old2get_annotations(infilename: FileName, patterns: Tuple[Pattern, Pattern]) \
#         -> Tuple[UttWordDict, Dict[Tuple[Level, Item], List[Tuple[UttId, Position]]]]:
#     '''
#     Reads the file with name filename in SASTA Annotation Format
#     :param infilename:
#     :param patterns
#     :return: a dictionary  with as  key a tuple (level, item) and as value a list of (uttid, tokenposition) pairs
#     '''
#
#     thedata = defaultdict(list)
#     #cdata = {}
#
#     allutts = {}
#
#     # To open Workbook
#     wb = xlrd.open_workbook(infilename)
#     sheet = wb.sheet_by_index(0)
#
#     startrow = 0
#     startcol = 0
#     headerrow = 0
#     headers = {}
#     lastrow = sheet.nrows
#     lastcol = sheet.ncols
#     #    firstwordcol = 2
#     #    lastwordcol = lastcol - 4
#     levelcol = 1
#     uttidcol = 0
#     stagescol = -1
#     commentscol = -1
#
#     uttlevel = 'utt'
#
#     uttcount = 0
#
#     for rowctr in range(startrow, lastrow):
#         if rowctr == headerrow:
#             for colctr in range(startcol, lastcol):
#                 headers[colctr] = sheet.cell_value(rowctr, colctr)
#                 if iswordcolumn(headers[colctr]):
#                     lastwordcol = colctr
#                     if isfirstwordcolumn(headers[colctr]):
#                         firstwordcol = colctr
#                 elif clean(headers[colctr]) in speakerheaders:
#                     spkcol = colctr
#                 elif clean(headers[colctr]) in uttidheaders:
#                     uttidcol = colctr
#                 elif clean(headers[colctr]) in levelheaders:
#                     levelcol = colctr
#                 elif clean(headers[colctr]) in stagesheaders:
#                     stagescol = colctr
#                 elif clean(headers[colctr]) in commentsheaders:
#                     commentscol = colctr
#         else:
#             if sheet.cell_value(rowctr, uttidcol) != "":
#                 uttid = str(int(sheet.cell_value(rowctr, uttidcol)))
#             thelevel = sheet.cell_value(rowctr, levelcol)
#             thelevel = clean(thelevel)
#             all_levels.add(thelevel)
#             # if thelevel == uttlevel:
#             #    uttcount += 1
#             curuttwlist = []
#             for colctr in range(firstwordcol, sheet.ncols):
#                 if thelevel == uttlevel:
#                     curcellval = sheet.cell_value(rowctr, colctr)
#                     if curcellval != '':
#                         curuttwlist.append(curcellval)
#                 elif thelevel in literallevels and colctr != stagescol and colctr != commentscol:
#                     thelabel = sheet.cell_value(rowctr, colctr)
#                     if colctr > lastwordcol:
#                         tokenposition = 0
#                     else:
#                         tokenposition = colctr - firstwordcol + 1
#                     # thedata[(thelevel, thelabel)].append(uttid)
#                     cleanlevel = thelevel
#                     cleanlabel = thelabel
#                     if cleanlabel != '':
#                         thedata[(cleanlevel, cleanlabel)].append((uttid, tokenposition))
#                 elif thelevel != uttlevel and colctr != stagescol and colctr != commentscol:
#                     thelabelstr = sheet.cell_value(rowctr, colctr)
#                     thelevel = sheet.cell_value(rowctr, levelcol)
#                     if lastwordcol + 1 <= colctr < sheet.ncols:
#                         # prefix = headers[colctr] aangepast om het simpeler te houden
#                         prefix = ""
#                     else:
#                         prefix = ""
#                     cleanlevelsandlabels = getcleanlevelsandlabels(thelabelstr, thelevel, prefix, patterns)
#                     if colctr > lastwordcol:
#                         tokenposition = 0
#                     else:
#                         tokenposition = colctr - firstwordcol + 1
#                     for (cleanlevel, cleanlabel) in cleanlevelsandlabels:
#                         thedata[(cleanlevel, cleanlabel)].append((uttid, tokenposition))
#             if curuttwlist != []:
#                 allutts[uttid] = curuttwlist
#     # wb.close() there is no way to close the workbook
#     return allutts, thedata


def get_annotations(infilename: FileName, patterns: Tuple[Pattern, Pattern]) \
        -> Tuple[UttWordDict, Dict[Tuple[Level, Item], List[Tuple[UttId, Position]]]]:
    '''
    Reads the file with name filename in SASTA Annotation Format
    :param infilename:
    :param patterns
    :return: a dictionary  with as  key a tuple (level, item) and as value a list of (uttid, tokenposition) pairs
    '''

    thedata = defaultdict(list)

    allutts = {}

    # To open Workbook
    header, data = xlsx.getxlsxdata(infilename)

    levelcol = 1
    uttidcol = 0
    stagescol = -1
    commentscol = -1
    unalignedcol = -1

    uttlevel = 'utt'

    uttcount = 0

    for col, val in enumerate(header):
        if iswordcolumn(val):
            lastwordcol = col
            if isfirstwordcolumn(val):
                firstwordcol = col
        elif clean(val) in speakerheaders:
            spkcol = col
        elif clean(val) in uttidheaders:
            uttidcol = col
        elif clean(val) in levelheaders:
            levelcol = col
        elif clean(val) in stagesheaders:
            stagescol = col
        elif clean(val) in commentsheaders:
            commentscol = col
        elif clean(val) in unalignedheaders:
            unalignedcol = col
        else:
            pass  # maybe warn here that an unknow column header has been encountered?

    for row in data:
        if row[uttidcol] != "":
            uttid = str(int(row[uttidcol]))
        thelevel = row[levelcol]
        thelevel = clean(thelevel)
        all_levels.add(thelevel)
        # if thelevel == uttlevel:
        #    uttcount += 1
        curuttwlist = []
        for colctr in range(firstwordcol, len(row)):
            if thelevel == uttlevel:
                curcellval = row[colctr]
                if curcellval != '':
                    curuttwlist.append(curcellval)
            elif thelevel in literallevels and colctr != stagescol and colctr != commentscol:
                thelabel = row[colctr]
                if colctr > lastwordcol:
                    tokenposition = 0
                else:
                    tokenposition = colctr - firstwordcol + 1
                cleanlevel = thelevel
                cleanlabel = thelabel
                if cleanlabel != '':
                    thedata[(cleanlevel, cleanlabel)].append((uttid, tokenposition))
            elif thelevel != uttlevel and colctr != stagescol and colctr != commentscol:
                thelabelstr = row[colctr]
                thelevel = row[levelcol]
                if colctr == unalignedcol:
                    prefix = ''
                if lastwordcol + 1 <= colctr < len(row):
                    # prefix = headers[colctr] aangepast om het simpeler te houden
                    prefix = ""
                else:
                    prefix = ""
                cleanlevelsandlabels = getcleanlevelsandlabels(thelabelstr, thelevel, prefix, patterns)
                if colctr > lastwordcol or colctr == unalignedcol:
                    tokenposition = 0
                else:
                    tokenposition = colctr - firstwordcol + 1
                for (cleanlevel, cleanlabel) in cleanlevelsandlabels:
                    thedata[(cleanlevel, cleanlabel)].append((uttid, tokenposition))
        if curuttwlist != []:
            allutts[uttid] = curuttwlist
    return allutts, thedata


def update(thedict: Dict[QId, Tuple[Level, Item, List[Tuple[UttId, Position]]]], qid: QId,
           goldtuple: Tuple[Level, Item, List[Tuple[UttId, Position]]]):
    (level, item, thecounter) = goldtuple
    if qid in thedict:
        (oldlevel, olditem, oldcounter) = thedict[qid]
        thedict[qid] = (oldlevel, olditem, oldcounter + thecounter)
    else:
        thedict[qid] = goldtuple


def oldgetitem2levelmap(mapping: Dict[Tuple[Item, Level], Any]) -> Dict[Item, List[Level]]:
    resultmap: Dict[Item, List[Level]] = {}
    for (item, level) in mapping:
        if item in resultmap:
            resultmap[item].append(level)
        else:
            resultmap[item] = [level]
    return resultmap


def getitem2levelmap(mapping: Dict[Tuple[Item, Level], Any]) -> Dict[Item, Level]:
    resultmap: Dict[Item, Level] = {}
    for (item, level) in mapping:
        if item in resultmap:
            settings.LOGGER.error('Duplicate level {} for item {} with level {} ignored'.format(level, item, resultmap[item]))
        else:
            resultmap[item] = level
    return resultmap


def codeadapt(c: str) -> str:
    result = c
    result = re.sub(r'\.', r'\\.', result)
    result = re.sub(r'\(', r'\\(', result)
    result = re.sub(r'\)', r'\\)', result)
    result = re.sub(r'\?', r'\\?', result)
    result = re.sub(r'\*', r'\\*', result)
    result = re.sub(r'\+', r'\\+', result)
    result = re.sub(r' ', r'\\s+', result)
    return result


def mkpatterns(allcodes: List[str]) -> Tuple[Pattern, Pattern]:
    basepattern = r''
    sortedallcodes = sorted(allcodes, key=len, reverse=True)
    adaptedcodes = [codeadapt(c) for c in sortedallcodes]
    basepattern = r'' + '|'.join(adaptedcodes) + '|' + itemseppattern
    fullpattern = r'^(' + basepattern + r')*$'
    return (re.compile(basepattern), re.compile(fullpattern))


def get_golddata(filename: FileName, mapping: Dict[Tuple[Item, Level], QId], altcodes: Dict[Tuple[Item, Level], Tuple[Item, Level]],
                 queries: QueryDict, includeimplies: bool = False) \
        -> Tuple[UttWordDict, Dict[QId, Tuple[Level, Item, List[Tuple[UttId, Position]]]]]:
    # item2levelmap = {}
    mappingitem2levelmap = getitem2levelmap(mapping)
    altcodesitem2levelmap = getitem2levelmap(altcodes)
    allmappingitems = [item for (item, _) in mapping]
    allaltcodesitems = [item for (item, _) in altcodes]
    allitems = allmappingitems + allaltcodesitems
    patterns = mkpatterns(allitems)
    allutts, basicdata = get_annotations(filename, patterns)
    results: Dict[QId, Tuple[Level, Item, List[Tuple[UttId, Position]]]] = {}
    for thelevel, theitem in basicdata:
        thecounter = basicdata[(thelevel, theitem)]
        # unclear why this below here is needed
#        if (theitem, thelevel) in mapping:
#            mappingitem = theitem
#        elif (varitem, thelevel) in mapping:
#            mappingitem = varitem
#        else:
#            mappingitem = theitem
        if thelevel in literallevels:
            # we still have to determine how to deal with this
            pass
        elif (theitem, thelevel) in mapping:
            qid = mapping[(theitem, thelevel)]
            update(results, qid, (thelevel, theitem, thecounter))
            if includeimplies:
                for implieditem in queries[qid].implies:
                    impliedlevel = mappingitem2levelmap[implieditem]
                    if (implieditem, impliedlevel) in mapping:
                        impliedqid = mapping[(implieditem, impliedlevel)]
                        update(results, impliedqid, (impliedlevel, implieditem, thecounter))
                    else:
                        settings.LOGGER.error('Implied Item ({},{}) not found in mapping'.format(implieditem, impliedlevel))
        elif (theitem, thelevel) in altcodes:
            (altitem, altlevel) = altcodes[(theitem, thelevel)]
            qid = mapping[(altitem, altlevel)]
            update(results, qid, (altlevel, altitem, thecounter))
            settings.LOGGER.info(
                '{} of level {} invalid code replaced by {} of level {}'.format(theitem, thelevel, altitem, altlevel))
            if includeimplies:
                for implieditem in queries[qid].implies:
                    impliedlevel = mappingitem2levelmap[implieditem]
                    if (implieditem, impliedlevel) in mapping:
                        impliedqid = mapping[(implieditem, impliedlevel)]
                        update(results, impliedqid, (impliedlevel, implieditem, thecounter))
                    else:
                        settings.LOGGER.error('Implied Item ({},{}) not found in mapping'.format(implieditem, impliedlevel))
        elif theitem in mappingitem2levelmap:   # valid item but wrong level
            thecorrectlevel = mappingitem2levelmap[theitem]
            qid = mapping[(theitem, thecorrectlevel)]
            update(results, qid, (thecorrectlevel, theitem, thecounter))
            settings.LOGGER.info(
                'level {} of item {} replaced by correct level {}'.format(thelevel, theitem, thecorrectlevel))
            if includeimplies:
                for implieditem in queries[qid].implies:
                    impliedlevel = mappingitem2levelmap[implieditem]
                    if (implieditem, impliedlevel) in mapping:
                        impliedqid = mapping[(implieditem, impliedlevel)]
                        update(results, impliedqid, (impliedlevel, implieditem, thecounter))
                    else:
                        settings.LOGGER.error('Implied Item ({},{}) not found in mapping'.format(implieditem, impliedlevel))
        elif theitem in altcodesitem2levelmap:  # valid alternative item but wrong level
            theitemlevel = altcodesitem2levelmap[theitem]
            (thecorrectitem, thecorrectlevel) = altcodes[(theitem, theitemlevel)]
            qid = mapping[(thecorrectitem, thecorrectlevel)]
            update(results, qid, (thecorrectlevel, thecorrectitem, thecounter))
            settings.LOGGER.info(
                'level {} of item {} replaced by correct level {} and item {}'.format(
                    thelevel, theitem, thecorrectlevel, thecorrectitem)
            )
            if includeimplies:
                for implieditem in queries[qid].implies:
                    impliedlevel = mappingitem2levelmap[implieditem]
                    if (implieditem, impliedlevel) in mapping:
                        impliedqid = mapping[(implieditem, impliedlevel)]
                        update(results, impliedqid, (impliedlevel, implieditem, thecounter))
                    else:
                        settings.LOGGER.error('Implied Item ({},{}) not found in mapping'.format(implieditem, thecorrectlevel))

        else:
            settings.LOGGER.error('{} of level {} not a valid coding'.format(theitem, thelevel))
    return allutts, results


def exact2global(thedata: Dict[Tuple[Level, Item], List[Tuple[UttId, Position]]]) -> Dict[Tuple[Level, Item], Counter]:
    '''
    turns a dictionary with  as values a list of (uttid, pos) tuples into a dictionary with the same keys and as values a counter of uttid
    :param thedata:
    :return:
    '''

    cdata = {}
    for atuple in thedata:
        newvalue = [uttid for (uttid, _) in thedata[atuple]]
        cdata[atuple] = Counter(newvalue)
    return cdata


def richexact2global(thedata):
    '''
    turns a dictionary with  as values a tuple (level, item,list of (uttid, pos) tuples) into a dictionary with the
    same keys and as values a tuple (level, item, counter of uttid)
    :param thedata:
    :return:
    '''

    cdata = {}
    for thekey in thedata:
        (thelevel, theitem, exactlist) = thedata[thekey]
        newvalue = [uttid for (uttid, _) in exactlist]
        cdata[thekey] = (thelevel, theitem, Counter(newvalue))
    return cdata


def richscores2scores(richscores: Dict[QId, Tuple[Level, Item, Any]]) -> Dict[QId, Any]:
    scores = {}
    for queryid in richscores:
        scores[queryid] = richscores[queryid][2]
    return scores



from sastadev.lexicon import informlexicon
from sastadev.stringfunctions import accentaigu, delhyphenprefix


def test():

    assert accentaigu('aap') == ['aap', 'aáp', 'áap', 'ááp']
    assert delhyphenprefix('anti-bom', informlexicon) == []
    assert delhyphenprefix('ex-vrouw', informlexicon) == []
    assert delhyphenprefix('sergeant-majoor', informlexicon) == []
    assert delhyphenprefix('ver-verkoop', informlexicon) == ['verkoop']
    assert delhyphenprefix('vver-verkoop', informlexicon) == ['verkoop']
    assert delhyphenprefix('vvver-verkoop', informlexicon) == ['verkoop']


if __name__ == '__main__':
    test()



import re

c_as_s = r'c(?=[eiy])'
c_as_k = r'c(?=[^eiyh])'
qu_as_kw = r'qu'
y_as_i = r'y'
x_as_ks = r'x'
gh_as_g = r'gh'
eu_as_E = r'eu'
ei_as_J = r'ei'

#: The varibale *replacementpattenrs* contains replacements to deal with certain
#: alternative ways to spell a sound, e.g *c* is replaced by *s* in certain context,
#: by *k* in other contexts, etc.
replacementpatterns = [(y_as_i, 'ie'), (c_as_s, 's'), (c_as_k, 'k'), (qu_as_kw, 'kw'), (x_as_ks, 'ks'), (gh_as_g, 'g'),
                       (eu_as_E, 'E'), (ei_as_J, 'J')]
replacements = [(re.compile(pat), repl) for pat, repl in replacementpatterns]


def phoneticise(instr: str) -> str:
    '''
    The function *phoneticise* carries out substitutions for string patterns as given
    in the variable *replacements* derived from the variable *replacementpatterns*.

    .. autodata:: phonetics::replacementpatterns

    '''
    result = instr
    for regex, repl in replacements:
        result = regex.sub(repl, result)
    return result



import os

import pytest
from lxml import etree

from sastadev.treebankfunctions import getstree, indextransform

bareindexnodexpath = './/node[@index and not(@cat) and not(@word)]'

streestrings = {}
streestrings[1] = """
<alpino_ds version="1.3" id="dpc-vla-001171-nl-sen.p.114.s.7">
  <node begin="0" cat="top" end="20" id="0" rel="top">
    <node begin="13" end="14" id="1" lemma="," pos="punct" postag="LET()" pt="let" rel="--" root="," word=","/>
    <node begin="0" cat="smain" end="19" id="2" rel="--">
      <node begin="0" cat="np" end="13" id="3" index="1" rel="su" highlight="yes">
        <node begin="0" end="1" id="4" lemma="het" lwtype="bep" naamval="stan" npagr="evon" pos="det" postag="LID(bep,stan,evon)" pt="lid" rel="det" root="het" word="Het" highlight="yes"/>
        <node begin="1" buiging="met-e" end="2" graad="basis" id="5" lemma="blauw-groen" naamval="stan" pos="adj" positie="prenom" postag="ADJ(prenom,basis,met-e,stan)" pt="adj" rel="mod" root="blauw_groen" word="blauw-groene" highlight="yes"/>
        <node begin="2" end="3" genus="onz" getal="ev" graad="basis" id="6" lemma="logo" naamval="stan" ntype="soort" pos="noun" postag="N(soort,ev,basis,onz,stan)" pt="n" rel="hd" root="logo" word="logo" highlight="yes"/>
        <node begin="3" cat="pp" end="6" id="7" rel="mod" highlight="yes">
          <node begin="3" end="4" id="8" lemma="met" pos="prep" postag="VZ(init)" pt="vz" rel="hd" root="met" vztype="init" word="met" highlight="yes"/>
          <node begin="4" cat="np" end="6" id="9" rel="obj1" highlight="yes">
            <node begin="4" end="5" id="10" lemma="de" lwtype="bep" naamval="stan" npagr="rest" pos="det" postag="LID(bep,stan,rest)" pt="lid" rel="det" root="de" word="de" highlight="yes"/>
            <node begin="5" end="6" id="11" lemma="N" pos="noun" postag="SPEC(meta)" pt="spec" rel="hd" root="N" spectype="meta" word="N" highlight="yes"/>
          </node>
        </node>
        <node begin="6" cat="rel" end="13" id="12" rel="mod" highlight="yes">
          <node begin="6" end="7" getal="ev" id="13" index="2" lemma="dat" naamval="stan" pdtype="pron" persoon="3" pos="pron" postag="VNW(betr,pron,stan,vol,3,ev)" pt="vnw" rel="rhd" root="dat" status="vol" vwtype="betr" word="dat" highlight="yes"/>
          <node begin="6" cat="ssub" end="13" id="14" rel="body" highlight="yes">
            <node begin="6" end="7" id="15" index="2" rel="obj1" highlight="yes"/>
            <node begin="7" cat="np" end="12" id="16" rel="su" highlight="yes">
              <node begin="7" buiging="met-e" end="8" id="17" lemma="al" naamval="stan" npagr="agr" pdtype="det" pos="det" positie="prenom" postag="VNW(onbep,det,stan,prenom,met-e,agr)" pt="vnw" rel="det" root="alle" vwtype="onbep" word="alle" highlight="yes"/>
              <node begin="8" buiging="met-e" end="9" graad="basis" id="18" lemma="Nederlandstalig" naamval="stan" pos="adj" positie="prenom" postag="ADJ(prenom,basis,met-e,stan)" pt="adj" rel="mod" root="Nederlandstalig" word="Nederlandstalige" highlight="yes"/>
              <node begin="9" end="10" getal="mv" graad="basis" id="19" lemma="school" ntype="soort" pos="noun" postag="N(soort,mv,basis)" pt="n" rel="hd" root="school" word="scholen" highlight="yes"/>
              <node begin="10" cat="pp" end="12" id="20" rel="mod" highlight="yes">
                <node begin="10" end="11" id="21" lemma="in" pos="prep" postag="VZ(init)" pt="vz" rel="hd" root="in" vztype="init" word="in" highlight="yes"/>
                <node begin="11" end="12" genus="onz" getal="ev" graad="basis" id="22" lemma="Brussel" naamval="stan" ntype="eigen" pos="name" postag="N(eigen,ev,basis,onz,stan)" pt="n" rel="obj1" root="Brussel" word="Brussel" highlight="yes"/>
              </node>
            </node>
            <node begin="12" end="13" id="23" lemma="dragen" pos="verb" postag="WW(pv,tgw,mv)" pt="ww" pvagr="mv" pvtijd="tgw" rel="hd" root="draag" word="dragen" wvorm="pv" highlight="yes"/>
          </node>
        </node>
      </node>
      <node begin="14" end="15" id="24" lemma="zijn" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="ben" word="is" wvorm="pv"/>
      <node begin="0" cat="ppart" end="19" id="25" rel="vc">
        <node begin="0" end="13" id="26" index="1" rel="su"/>
        <node begin="15" buiging="zonder" end="16" id="27" lemma="uitgroeien" pos="verb" positie="vrij" postag="WW(vd,vrij,zonder)" pt="ww" rel="hd" root="groei_uit" word="uitgegroeid" wvorm="vd"/>
        <node begin="16" cat="pp" end="19" id="28" rel="pc">
          <node begin="16" end="17" id="29" lemma="tot" pos="prep" postag="VZ(init)" pt="vz" rel="hd" root="tot" vztype="init" word="tot"/>
          <node begin="17" cat="np" end="19" id="30" rel="obj1">
            <node begin="17" end="18" id="31" lemma="een" lwtype="onbep" naamval="stan" npagr="agr" pos="det" postag="LID(onbep,stan,agr)" pt="lid" rel="det" root="een" word="een"/>
            <node begin="18" end="19" genus="onz" getal="ev" graad="basis" id="32" lemma="kwaliteitslabel" naamval="stan" ntype="soort" pos="noun" postag="N(soort,ev,basis,onz,stan)" pt="n" rel="hd" root="kwaliteitslabel" word="kwaliteitslabel"/>
          </node>
        </node>
      </node>
    </node>
    <node begin="19" end="20" id="33" lemma="." pos="punct" postag="LET()" pt="let" rel="--" root="." word="."/>
  </node>
  <sentence>Het blauw-groene logo met de N dat alle Nederlandstalige scholen in Brussel dragen , is uitgegroeid tot een kwaliteitslabel .</sentence>
  <comments>
    <comment>Q#dpc-vla-001171-nl-sen.p.114.s.7|Het blauw-groene logo met de N dat alle Nederlandstalige scholen in Brussel dragen , is uitgegroeid tot een kwaliteitslabel .|</comment>
  </comments>
</alpino_ds>
"""

strees = {i: etree.fromstring(streestrings[i]) for i in streestrings}


@pytest.mark.skip(reason='test code does not work')
def test():
    for i in strees:
        idxnodes = strees[i].xpath(bareindexnodexpath)
        if idxnodes == []:
            print('No bare index nodes in tree {i}')
        # newtree = newindextransform(strees[i])
        newtree = indextransform(strees[i])
        etree.dump(newtree)


def testwholelassy():
    lassykleinpath = r'D:\Dropbox\various\Resources\LASSY\LASSY-Klein\Lassy-Klein\Treebank'
    for root, dirs, files in os.walk(lassykleinpath):
        print(f'Processing {root}...')
        for filename in files:
            base, ext = os.path.splitext(filename)
            if ext == '.xml':
                fullname = os.path.join(root, filename)
                fullstree = getstree(fullname)
                stree = fullstree.getroot()
                expansion = indextransform(stree)
                if expansion.xpath(bareindexnodexpath) != []:
                    print(fullname)
                    etree.dump(stree)
                    etree.dump(expansion)


def main():
    test()
    # testwholelassy()


if __name__ == '__main__':
    main()



import os

from sastadev.conf import settings
from sastadev.constants import intreebanksfolder, outtreebanksfolder
from sastadev.dedup import filledpauseslexicon
from sastadev.treebankfunctions import (getattval, getnodeyield, getstree,
                                        getuttid, getyield, onbvnwdet)

digits2 = r'\d\d'

asta = 'asta'
tarsp = 'tarsp'
stap = 'stap'
dld = 'dld'
schl = 'schlichting'
auris = 'auris'


def shorten(fullname):
    _, filename = os.path.split(fullname)
    basename, _ = os.path.splitext(filename)
    return basename


def getpaths(dataset):
    result = [os.path.join(settings.DATAROOT, 'VKLStap', intreebanksfolder),
              os.path.join(settings.DATAROOT, 'VKLStap', outtreebanksfolder)]
    return result


paths = {}

paths[stap] = getpaths('VKLStap')
paths[asta] = getpaths('VKLASTA')
paths[tarsp] = getpaths('VKLTarsp')
paths[dld] = getpaths('Auris')

ext = '.xml'


class Ngram:
    def __init__(self, n, cond):
        self.n = n
        self.cond = cond


def commonancestor(node1, node2, stree, ancestorcond):
    nodes = stree.xpath('.//node')
    ancestorcandidates = [n for n in nodes if ancestorcond(n)]
    for anc in ancestorcandidates:
        descendants = anc.xpath('.//node')
        if node1 in descendants and node2 in descendants:
            return True
    return False


def lemma(x):
    return getattval(x, 'lemma')


def word(x):
    return getattval(x, 'word')


# det = lambda x : pt(x) == 'lw' or (pt(x) == 'vnw' and getattval(x, 'vwtype') in {'aanw', 'bez'})

def pt(x):
    return getattval(x, 'pt')


def findmatches(ngram, leaves):
    matches = []
    for i in range(len(leaves) - ngram.n + 1):
        if ngram.cond(leaves[i:i + ngram.n], leaves, i):
            newmatch = (i, i + ngram.n)
            matches.append(newmatch)
    return matches


def getfilenames(ds, session=None):
    inpath = paths[ds]
    if session is not None:
        def cond(x): return (x + 1 == session)
    else:
        def cond(_): return True
    infullnames = [os.path.join(path, ifn) for path in paths[ds] for ifn in os.listdir(path)]
    return infullnames


def Rpronoun(el):
    result = lemma(el) in {'er', 'hier', 'daar'}
    return result


def ispv(n):
    result = pt(n) == 'ww' and getattval(n, 'wvorm') == 'pv'
    return result


def sipvjpvjsi(ns, stree):
    resultmatches = []
    ngram1 = Ngram(2, cond13)
    matches = findmatches(ngram1, ns)
    for (b, e) in matches:
        vnw1 = ns[b]
        pv1 = ns[e - 1]
        pv2found = False
        for mc in range(e, len(ns) - 2):
            if ispv(ns[mc]):
                if word(ns[mc]).lower() == word(pv1).lower():
                    pv2 = ns[mc]
                    if not commonancestor(pv1, pv2, stree, lambda x: getattval(x, 'cat') == 'smain'):
                        pv2found = True
                        pv2pos = mc
                    break
            else:
                mc += 1
        if pv2found:
            vnw2found = word(ns[pv2pos + 1]).lower() == word(vnw1).lower()
            if vnw2found:
                resultmatch = (b, e, mc, mc + 2)
                resultmatches.append(resultmatch)
    return resultmatches


def cond17f(ns, lvs, i):
    result = lemma(ns[0]) == 'te'
    result = result and getattval(ns[1], 'his') == 'te_verb'
    result = result and lemma(ns[2]) == 'te'
    return result


def cond1(ns, _, i): return pt(ns[0]) == 'vnw' and pt(ns[1]) == 'ww' and getattval(ns[1], 'wvorm') == 'pv' and \
    pt(ns[2]) == 'vnw' and lemma(ns[0]) == lemma(ns[2]) and pt(ns[3]) == 'ww' and \
    getattval(ns[3], 'wvorm') == 'pv'


def det(el): return (pt(el) == 'lid'
                     or (pt(el) == 'vnw' and getattval(el, 'vwtype') in {'aanw', 'bez'})
                     or (pt(el) == 'vnw' and getattval(el, 'vwtype') == 'onbep' and onbvnwdet(el))
                     ) and \
    not Rpronoun(el)   # and getattval(el, 'positie') == 'prenom' left out


def erhierdaar(el): return lemma(el) in {'er', 'hier', 'daar'}


def cond2(ns, _, i): return pt(ns[0]) == 'vz' and det(ns[1]) and (not erhierdaar(ns[1])) and pt(ns[2]) == 'vz' and \
    det(ns[3]) and (not erhierdaar(ns[3]))


def cond3(ns, _, i): return pt(ns[0]) == 'n' and pt(ns[1]) == 'n'
def cond4(ns, lvs, i): return lemma(ns[0]) == "," and ns[1] == lvs[-1]
def cond4a(ns, lvs, i): return lemma(ns[0]) == "," and pt(ns[2]) == 'let' and ns[2] == lvs[-1]
def cond5(ns, _, i): return lemma(ns[0]) == 'nou' and lemma(ns[1]) == 'ja'
def cond6(ns, lvs, i): return pt(ns[0]) == 'n' and len([lv for lv in lvs[:i - 1] if lemma(lv) == lemma(ns[0])]) > 1
def cond7(ns, lvs, i): return pt(ns[0]) == 'vg' and det(ns[1]) and lemma(ns[2]) == lemma(ns[0]) and det(ns[3])
def cond8(ns, lvs, i): return word(ns[2]).startswith(word(ns[0])) and word(ns[3]).startswith(word(ns[1]))


def cond9(ns, _, i): return pt(ns[0]) == 'vnw' and getattval(ns[0], 'vwtype') == 'pers' and pt(ns[1]) == 'ww' and \
    getattval(ns[1], 'wvorm') == 'pv' and pt(ns[2]) == 'vnw' and pt(ns[3]) == 'ww' and \
    getattval(ns[3], 'wvorm') == 'pv'  # and \
# (word(ns[0]) == word(ns[2]) or word(ns[1]) == word(ns[3]) or lemma(ns[1]) == lemma(ns[3]))


def cond10(ns, lvs, i): return pt(ns[0]) == 'vnw' and getattval(ns[0], 'vwtype') == 'pers' and \
    pt(ns[1]) == 'ww' and getattval(ns[1], 'wvorm') == 'pv' and \
    pt(ns[2]) != 'ww' and word(ns[3]) == word(ns[1]) and word(ns[4]) == word(ns[0])


def cond11(ns, lvs, i): return lemma(ns[0]) == lemma(ns[2]) and lemma(ns[1]) == lemma(ns[3]) and \
    (word(ns[0]) != word(ns[2]) or word(ns[1]) != word(ns[3]))


def cond12(ns, lvs, i): return pt(ns[0]) == 'vz' and det(ns[1]) and pt(ns[2]) == 'vz' and det(ns[3])  # == cond2


def cond13(ns, lvs, i): return pt(ns[0]) == 'vnw' and getattval(ns[0], 'vwtype') == 'pers' and pt(ns[1]) == 'ww' and \
    getattval(ns[1], 'wvorm') == 'pv'


def cond14(ns, lvs, i): return ns[0] != lvs[0] and lemma(ns[0]) != 'nou' and (lemma(ns[1]) in filledpauseslexicon or pt(ns[0]) == 'tsw') and lemma(ns[2]) == 'nee'
def cond15(ns, lvs, i): return ns[0] != lvs[0] and lemma(ns[1]) == 'nee'


def cond16(ns, lvs, i): return det(ns[0]) and det(ns[2]) and lemma(ns[1]) == lemma(ns[3]) \
    and pt(ns[1]) == 'n' and pt(ns[3]) == 'n'
def cond16a(ns, lvs, i): return lemma(ns[0]) in {'een', 'geen'} and lemma(ns[2]) in {'een', 'geen'} and lemma(ns[1]) == lemma(ns[3]) \
    and pt(ns[1]) == 'n' and pt(ns[3]) == 'n'


def cond17(ns, lvs, i): return lemma(ns[0]) == 'te' and getattval(ns[1], 'his') == 'te_verb' and lemma(ns[2]) == 'te'
# cond17 = cond17f
def cond17a(ns, lvs, i): return lemma(ns[0]) == 'te' and word(ns[1]) == 'kregen' and lemma(ns[2]) == 'te'


def cond18(ns, lvs, i): return pt(ns[0]) == 'vz' and lemma(ns[1]) in {'dit', 'dat', 'deze', 'die'}


ngram1 = Ngram(4, cond1)
ngram2 = Ngram(4, cond2)
ngram3 = Ngram(2, cond3)
ngram4 = Ngram(2, cond4)  # uitloop maar zonder punc daarachter
ngram4a = Ngram(3, cond4a)  # uitlopp maar met punc daarachter
ngram5 = Ngram(2, cond5)
ngram6 = Ngram(1, cond6)
ngram7 = Ngram(4, cond7)
ngram8 = Ngram(4, cond8)  # partial repeat sequence
ngram9 = Ngram(4, cond9)  # vnw pv vnw pv
ngram10 = Ngram(5, cond10)  # vnwj pvi x pvi vnwj
ngram11 = Ngram(4, cond11)  # lemma_i lemma_j lemma_i lemma_j
ngram12 = Ngram(4, cond12)  # vz det vz det
ngram14 = Ngram(3, cond14)  # fp/tsw + nee , to apply to the uncleaned leaves
ngram15 = Ngram(2, cond15)  # word + nee, to apply to the cleanleaves
ngram16 = Ngram(4, cond16)  # geen beroerte een beroerte
ngram16a = Ngram(4, cond16a)  # geen beroerte een beroerte test
ngram17 = Ngram(4, cond17)  # te kregen te krijgen
ngram17a = Ngram(4, cond17a)  # te kregen te krijgen test
ngram18 = Ngram(2, cond18)  # met dit


def main():

    infullnames = []
    for ds in [tarsp, asta, stap, dld]:
        infullnames += getfilenames(ds)

    # for ds in [asta]:
    #    infullnames += getfilenames(ds, session=4)

    for infullname in infullnames:
        short = shorten(infullname)
        fulltreebank = getstree(infullname)
        if fulltreebank is not None:
            treebank = fulltreebank.getroot()
            for tree in treebank:
                uttid = getuttid(tree)
                leaves = getnodeyield(tree)
                cleanleaves = [leave for leave in leaves if getattval(leave, 'word') not in filledpauseslexicon]
                cleanwordlist = [getattval(leave, 'word') for leave in cleanleaves]
                matches = findmatches(ngram18, cleanleaves)
                # matches = sipvjpvjsi(cleanleaves, tree)
                for match in matches:
                    uttid = getuttid(tree)
                    cleanleaves_str = [word(el) for el in cleanleaves]
                    print(short, uttid, match, cleanleaves_str, getyield(tree))


if __name__ == '__main__':
    main()



import pathlib
from setuptools import setup

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")


setup(
    name='sastadev',
    version='0.1.0',
    description='Linguistic functions for SASTA tool',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/UUDigitalHumanitieslab/sastadev',
    author='Research Software Lab - Centre for Digital Humanities - Utrecht University',
    author_email='digitalhumanities@uu.nl',
    package_dir={'': 'src'},
    packages=['sastadev'],
    python_requires='>=3.7, <4',
    package_data={'sastadev': ['py.typed', 'data/**/*']},
    license='BSD-3-Clause',
    entry_points={
        'console_scripts': [
            'sastadev=sastadev.__main__:main'
        ],
    },
    install_requires=[
        'auchann',
        'lxml',
        'openpyxl',
        'XlsxWriter'
    ],
)


'''
The module TARSPpostfunctions defines functions for the TARSP post part of the methods

'''
from collections import Counter
from typing import Dict, List

from sastadev.allresults import AllResults
from sastadev.conf import settings
from sastadev.query import core_process
from sastadev.sastatypes import QId, QueryDict, Stage, SynTree
from sastadev.treebankfunctions import getmeta

OndVC = 'T071'
OndWVC = 'T076'
OndWBVC = 'T075'

#: The variable (constant) *vuqueryids* contains a list of Query identifiers for
#: queries for fixed expressions (V.U.).
vuqueryids = ['T094', 'T095', 'T096', 'T150']
#: The variable (constant) *tarsp_clausetypes* contains the (lower case) values for the
#: subcategory of a query that represent clause types.
tarsp_clausetypes = ['mededelende zin', 'vragen', 'gebiedende wijs']
#: The variable (constant) *excludedqids* contains a list of QIds for queries that
#: should be excluded in computing G.O. Fase.
excludedqids = ['T039', 'T048', 'T049', 'T052']   # TARSP p. 21: hè, Into, Inversie, Kop
#: The variable (constant) *gofase_minthreshold* contains the value of the minimum
#: percentage of analysis units that must have been scored to be included in G.O. Fase.
gofase_minthreshold = 0.05  # 5% p21 Tarsp 2005


def getqueriesbystage(queries: QueryDict) -> Dict[Stage, List[QId]]:
    '''
    The function *getqueriesbystage* creates a dictionary with a stage as key and a
    list of QIds as value.

    It selects those QIds for which the query's (lower cased) subcategory is contained
    in the constant *tarsp_clausetypes*:

    .. autodata:: TARSPpostfunctions::tarsp_clausetypes

    and which is not included in the list of *excludedqids*:

    .. autodata:: TARSPpostfunctions::excludedqids

    if these conditions are met, the QId is appended to the dictionary item with key
    equal to the stage of the query associated with QId.
    '''
    results = {}
    for qid in queries:
        if queries[qid].subcat.lower() in tarsp_clausetypes and qid not in excludedqids:
            stage = queries[qid].fase
            if stage in results:
                results[stage].append(qid)
            else:
                results[stage] = [qid]
    return results


def vutotaal(allresults: AllResults, _: SynTree) -> int:
    '''
    The function *vutotaal* computes the total number of  "Vaste Uitdrukkingen" (VU) in
    the variable *allresults*. It uses the set *vuqueryids* to determine which
    queries to take into account and which not:

     .. autodata:: vuqueryids
    '''
    scores = []
    for qid in vuqueryids:
        if qid in allresults.coreresults:
            scores.append(allresults.coreresults[qid])
        else:
            scores.append(Counter())
#    scores = [allresults.coreresults[qid] for qid in vuqueryids]
    counts = [len(s) for s in scores]
    result = sum(counts)
    return result


def gtotaal(allresults: AllResults, _: SynTree) -> int:
    '''
    The function *gtotaal* computes the number of utterances to be analysed. It does
    so by subtracting the number of V.U. utterances and the results for *Atotaal* from
    the total number of utterances.
    '''
    Atotaal = 0
    vutotaal = allresults.postresults['T151']
    Gtotaal = allresults.uttcount - Atotaal - vutotaal
    return Gtotaal


def countutts(acounter: Counter):
    '''
    The function *countutts* returns the sum of the values for each key in
    *acounter*.
    '''
    result = 0
    for k in acounter:
        result += acounter[k]
    return result


def getuttcountsbystage(queriesbystage: Dict[Stage, List[QId]], allresults: AllResults)\
        -> Dict[Stage, int]:
    '''
    The function *getuttcountsbystage* computes a dictionary *uttcounts* of Stage,
    int items based on the input parameters *queriesbystage* and *allresults*.

    For each qid in *queriesbystage* that is a qid for a core query, it counts the
    number of utterances marked in  *allresults.coreresults*. For this it uses the
    function *countutts*:

    .. autofunction:: TARSPpostfunctions::countutts

    '''
    uttcounts = {}
    for stage in queriesbystage:
        uttcounts[stage] = 0
        for qid in queriesbystage[stage]:
            if qid in allresults.coreresults:
                uttcounts[stage] += countutts(allresults.coreresults[qid])
    return uttcounts


def getstage(uttcounts: Dict[Stage, int], allresults: AllResults) -> Stage:
    '''
    The function *getstage* computes the stage on the basis of the *uttcounts*
    dictionary with Stage, int items and *allresults*

    The stage is taken into consideration if its number of scores divided by *gtotaal* is
    greater or equal to the value of *gofase_minthreshold*:

    .. autodata:: TARSPpostfunctions::gofase_minthreshold

    From the remaining candidates the highest stage value is selected.
    '''
    cands = []
    gtotaal = allresults.postresults['T152']
    for el in uttcounts:
        if gtotaal != 0:
            if uttcounts[el] / gtotaal >= gofase_minthreshold:
                cands.append(el)
        else:
            settings.LOGGER.error('gtotaal has value 0')
    if cands == []:
        result = 1
    else:
        result = max(cands)
    return result


def gofase(allresults: AllResults, thequeries: QueryDict) -> Stage:
    '''
    The function *gofase* computes the stage given the results in the parameter
    *allresults* and the queries in the parameter *thequeries*.

    It first obtains *queriesbystage*, a dictionary of Stage, List[QId] items, via the
    function  *getqueriesbystage* applied to *thequeries*:

    .. autofunction:: TARSPpostfunctions::getqueriesbystage

    Next, it obtains *uttcounts*,  a dictionary of Stage, int items by applying the
    function  *getuttcountsbystage* to *queriesbystage* and *allresults*:

    .. autofunction:: TARSPpostfunctions::getuttcountsbystage

    Finally, it obtains the stage by applying the function *getstage* to *uttcounts*
    and *allresults*:

    .. autofunction:: TARSPpostfunctions::getstage

    and then it returns the obtained *stage*.
    '''
    result = 0
    queriesbystage: Dict[Stage, List[QId]] = getqueriesbystage(thequeries)
    uttcounts: Dict[Stage, int] = getuttcountsbystage(queriesbystage, allresults)
    result: Stage = getstage(uttcounts, allresults)

    return result


def genpfi(stage: Stage, allresults: AllResults, allqueries: QueryDict) -> int:
    '''
    The function *genpfi* computes the *Profielscore* (PF) for the stage given by the
    parameter *stage* on the basis of *allresults* and the query dictionary *allqueries*.
    It selects the queries of the given stage that are core queries and that are not
    *star2* queries.

    From these, it only selects the ones for which the number of results is larger than 0.
    It adds *OndVC* if *OndWVC* or *OndWBVC* has been scored.
    Special measures for *Xneg*, *OndB*, *VCW* and *BX* still have to be implemented.
    The description in Schlichting (p. 23) is not specific enough.
    '''
    theqids = [qid for qid in allqueries if allqueries[qid].fase == stage and allqueries[qid].process == core_process
               and allqueries[qid].special1 != 'star2']
    coreresults = allresults.coreresults
    scoredqids = [qid for qid in theqids if qid in coreresults and len(coreresults[qid]) > 0]
    # OndVC
    if OndWVC in theqids or OndWBVC in scoredqids:
        scoredqids.append(OndVC)
    # XNeg
    # OndB
    # VCW
    # BX
    result = len(scoredqids)
    return result


def pf2(allresults: AllResults, allqueries: QueryDict) -> int:
    '''
    The function *pf2* uses the function *genpfi* to compute the 'Profielscore' for Stage II
    '''
    return genpfi(2, allresults, allqueries)


def pf3(allresults: AllResults, allqueries: QueryDict) -> int:
    '''
    The function *pf3* uses the function *genpfi* to compute the 'Profielscore' for Stage III
    '''
    return genpfi(3, allresults, allqueries)


def pf4(allresults: AllResults, allqueries: QueryDict) -> int:
    '''
    The function *pf4* uses the function *genpfi* to compute the 'Profielscore' for Stage IV
    '''
    return genpfi(4, allresults, allqueries)


def pf5(allresults: AllResults, allqueries: QueryDict):
    '''
    The function *pf5* uses the function *genpfi* to compute the 'Profielscore' for Stage V
    '''
    return genpfi(5, allresults, allqueries)


def pf6(allresults: AllResults, allqueries: QueryDict) -> int:
    '''
    The function *pf6* uses the function *genpfi* to compute the 'Profielscore' for Stage VI
    '''
    return genpfi(6, allresults, allqueries)


def pf7(allresults: AllResults, allqueries: QueryDict) -> int:
    '''
    The function *pf7* uses the function *genpfi* to compute the 'Profielscore' for Stage VII
    '''
    return genpfi(7, allresults, allqueries)


def pf(allresults: AllResults, allqueries: QueryDict) -> int:
    '''
    The function *pf* computes the *'Profielscore'* for the whole sample (*PF*) by
    summing the  'Profielscore's per stage as computed by *pf2* through *pf7*.

    The *'Profielscore's* per stage are computed by *pf2* through *pf7*, each of which
    uses the function *genpfi*:

    .. autofunction:: TARSPpostfunctions::genpfi

    '''
    postresults = allresults.postresults
    result = sum([postresults['T154'], postresults['T155'], postresults['T158'],
                  postresults['T159'], postresults['T160'], postresults['T161']])
    return result


def getname(allresults: AllResults, allqueries: QueryDict) -> str:
    '''
    The function *getname* obtains the name of the patient/child being investigated
    from the metadata. It uses the function *getmeta* to achieve this.

    '''
    result = getmeta('name')
    return result


def getchildage(allresults: AllResults, allqueries: QueryDict) -> str:
    '''
    The function *getchildage* is intended to obtain the age of the child being
    investigated from the metadata. It still has to be implemented. Currently it simply returns the empty string..

    '''
    result = ''
    return result



import re

interpunctionpattern = r'[\[]]!"#$%&\'\(\)\*\+,\-\./^_`{|}~@=\:;<>?]'
ch_finalinterpunctionpattern = r'[\[\]!"#$%&\'\(\*\+,\-\./^_`{|}~@=\:;<>?]'  # ) left out
ch_initialinterpunctionpattern = r'[\[\]!"#$%\'\)\*\+,\-\./^_`{|}~@=\:;<>?]'  # ( and & left out

initpattern = r'^(' + ch_initialinterpunctionpattern + r')(.*)$'
initre = re.compile(initpattern)
finalpattern = r'^(.*)(' + ch_finalinterpunctionpattern + r')$'
finalre = re.compile(finalpattern)


def tokenise(str):
    #split by spaces
    toklist1 = str.split()
    #remove final interpunction
    toklist2 = []
    for tok in toklist1:
        curtokpart = tok
        finallist = []
        while True:
            m = finalre.match(curtokpart)
            if m is None:
                if curtokpart != '':
                    toklist2.append(curtokpart)
                toklist2 = toklist2 + finallist
                break
            else:
                finallist = [m.group(2)] + finallist
                curtokpart = m.group(1)

    #remove initital punctuation
    toklist3 = []
    for tok in toklist2:
        curtokpart = tok
        while True:
            m = initre.match(curtokpart)
            if m is None:
                if curtokpart != '':
                    toklist3.append(curtokpart)
                break
            else:
                toklist3.append(m.group(1))
                curtokpart = m.group(2)
    return toklist3



'''
This module defines the function read_method to read in a method:

* read_method(methodfilename: FileName) -> Tuple[QueryDict, Item_Level2QIdDict, AltCodeDict, List[QId]]:
'''

from typing import List, Tuple

from sastadev import xlsx
from sastadev.conf import settings
from sastadev.query import Query, form_process, post_process
from sastadev.sastatypes import (AltCodeDict, FileName, Item_Level2QIdDict,
                                 QId, QueryDict)

comma = ','

altitemsep = comma
implies_sep = comma
itemseppattern = r'[,-;\s]'


def getboolean(str: str) -> bool:
    if str is None:
        result = False
    elif str == '':
        result = False
    else:
        cleanstr = str.strip().lower()
        if cleanstr in ['no', 'false'] or cleanstr[0] in ['n', 'f']:
            result = False
        else:
            result = True
    return result


def getint(fase: str) -> int:
    try:
        result = int(fase)
    except Exception:
        result = 0
    return result


def get_pages(val: str) -> str:
    #pages = val.split(pagesep)
    #result = pages
    result = val
    return result


def getaltitems(str: str) -> List[str]:
    result = getlistofitems(str, altitemsep)
    return result


def getimplies(str: str) -> List[str]:
    result = getlistofitems(str, implies_sep)
    return result


def getlistofitems(str: str, sep: str) -> List[str]:
    rawresult = str.split(sep)
    cleanresult = [w.strip().lower() for w in rawresult]
    if cleanresult == ['']:
        cleanresult = []
    return cleanresult


# def oldread_method(methodfilename: FileName) -> Tuple[QueryDict, Item_Level2QIdDict, AltCodeDict, List[QId]]:
#     # To open Workbook
#     wb = xlrd.open_workbook(methodfilename)
#     sheet = wb.sheet_by_index(0)
#
#     idcol, catcol, subcatcol, levelcol, itemcol, altcol, impliescol, \
#         originalcol, pagescol, fasecol, querycol, informcol, screeningcol, processcol, special1col, special2col, commentscol = \
#         0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16
#
#     headerrow = 0
#
#     queries: QueryDict = {}
#     item2idmap: Item_Level2QIdDict = {}
#     altcodes: AltCodeDict = {}
#
#     postquerylist: List[QId] = []
#     for rowctr in range(sheet.nrows):
#         if rowctr != headerrow:
#             id : QId = sheet.cell_value(rowctr, idcol).strip()
#             cat: str = sheet.cell_value(rowctr, catcol).strip()
#             subcat: str = sheet.cell_value(rowctr, subcatcol).strip()
#             level: str = sheet.cell_value(rowctr, levelcol).strip()
#             item: str = sheet.cell_value(rowctr, itemcol).strip()
#             altitems: List[str] = getaltitems(sheet.cell_value(rowctr, altcol))
#             implies: List[str] = getimplies(sheet.cell_value(rowctr, impliescol))
#             original: bool = getboolean(sheet.cell_value(rowctr, originalcol))
#             pages: str = get_pages(sheet.cell_value(rowctr, pagescol))
#             fase: int = getint(sheet.cell_value(rowctr, fasecol))
#             query: str = sheet.cell_value(rowctr, querycol)
#             inform: str = sheet.cell_value(rowctr, informcol)
#             screening: str  = sheet.cell_value(rowctr, screeningcol)
#             process: str = sheet.cell_value(rowctr, processcol).strip()
#             special1: str = sheet.cell_value(rowctr, special1col).strip()
#             special2: str = sheet.cell_value(rowctr, special2col).strip()
#             comments: str = sheet.cell_value(rowctr, commentscol)
#
#             queries[id] = Query(id, cat, subcat, level, item, altitems, implies, original, pages, fase, query, inform, screening, process,
#                                 special1, special2, comments)
#             if queries[id].process in [post_process, form_process]:
#                 postquerylist.append(id)
#             lcitem = item.lower()
#             lclevel = level.lower()
#             if (lcitem, lclevel) in item2idmap:
#                 settings.LOGGER.error('Duplicate (item, level) pair for {} and {}'.format(item2idmap[(lcitem, lclevel)], id))
#             item2idmap[(lcitem, lclevel)] = id
#             for altitem in altitems:
#                 lcaltitem = altitem.lower()
#                 if (lcaltitem, lclevel) in altcodes:
#                     settings.LOGGER.error('Duplicate (alternative item, level) pair for {} and {}'.format(altcodes[(lcaltitem, lclevel)], id))
#                 altcodes[(lcaltitem, lclevel)] = (lcitem, lclevel)
#
#         rowctr += 1
#     return (queries, item2idmap, altcodes, postquerylist)

def empty(row: list) -> bool:
    for el in row:
        if el != '':
            return False
    return True


def read_method(methodfilename: FileName) -> Tuple[QueryDict, Item_Level2QIdDict, AltCodeDict, List[QId]]:
    header, data = xlsx.getxlsxdata(methodfilename)

    idcol, catcol, subcatcol, levelcol, itemcol, altcol, impliescol, \
        originalcol, pagescol, fasecol, querycol, informcol, screeningcol, processcol, special1col, special2col, commentscol = \
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16

    headerrow = 0

    queries: QueryDict = {}
    item2idmap: Item_Level2QIdDict = {}
    altcodes: AltCodeDict = {}

    postquerylist: List[QId] = []
    for row in data:
        if not empty(row):
            id: QId = row[idcol].strip()
            cat: str = row[catcol].strip()
            subcat: str = row[subcatcol].strip()
            level: str = row[levelcol].strip()
            item: str = row[itemcol].strip()
            altitems: List[str] = getaltitems(row[altcol])
            implies: List[str] = getimplies(row[impliescol])
            original: bool = getboolean(row[originalcol])
            pages: str = get_pages(row[pagescol])
            fase: int = getint(row[fasecol])
            query: str = row[querycol]
            inform: str = row[informcol]
            screening: str = row[screeningcol]
            process: str = row[processcol].strip()
            special1: str = row[special1col].strip()
            special2: str = row[special2col].strip()
            comments: str = row[commentscol]

            queries[id] = Query(id, cat, subcat, level, item, altitems, implies, original, pages, fase, query, inform, screening, process,
                                special1, special2, comments)
            if queries[id].process in [post_process, form_process]:
                postquerylist.append(id)
            lcitem = item.lower()
            lclevel = level.lower()
            if (lcitem, lclevel) in item2idmap:
                settings.LOGGER.error('Duplicate (item, level) pair for {} and {}'.format(item2idmap[(lcitem, lclevel)], id))
            item2idmap[(lcitem, lclevel)] = id
            for altitem in altitems:
                lcaltitem = altitem.lower()
                if (lcaltitem, lclevel) in altcodes:
                    settings.LOGGER.error('Duplicate (alternative item, level) pair for {} and {}'.format(altcodes[(lcaltitem, lclevel)], id))
                altcodes[(lcaltitem, lclevel)] = (lcitem, lclevel)

    return (queries, item2idmap, altcodes, postquerylist)



'''
The macros module:

* reads in macro files
* parses the macro files in the macros subfolder
* stores them in a macro expansion dictionary macrodict with macro: expansion items as  key:value items
* stores  macros generated by generatemacros.generatemacros in the macrodict dictionary
* provides the function expandmacros of type ``str -> str`` to convert a string into a different string by expanding the macros in the input string

'''

import os.path as op
import re
from typing import Dict, List, TextIO

from sastadev.conf import settings
from sastadev.generatemacros import generatemacros

idpat = r'([A-z_][A-z0-9_]*)'
eqpat = r'='
exprpat = r'"""(.*?)"""'
whitespaces = r'\s+'

macrocallpat = r'(%.+?%)'
macrocallre = re.compile(macrocallpat)

macropat = idpat + whitespaces + eqpat + whitespaces + exprpat

macrore = re.compile(macropat, re.S)


def macrostrs2dict(teststrings: List[str]) -> Dict[str, str]:
    macrodict = {}
    for tstr in teststrings:
        macromatches = macrore.finditer(tstr)
        for macromatch in macromatches:
            macroname = macromatch.group(1)
            macroexpr = macromatch.group(2)
            macrodict[macroname] = macroexpr

    return macrodict


def readmacros(macrofile: TextIO, macrodict: Dict[str, str]) -> Dict[str, str]:
    macrotext = macrofile.read()
    macromatches = macrore.finditer(macrotext)
    for macromatch in macromatches:
        macroname = macromatch.group(1)
        macroexpr = macromatch.group(2)
        if macroname in macrodict:
            settings.LOGGER.warning('Duplicate macro {} encountered. Ignored'.format(macroname))
        else:
            macrodict[macroname] = macroexpr
    return macrodict


def expandmacros(expr: str) -> str:
    result = expandmacrosdict(expr, macrodict)
    return result


def expandmacrosdict(expr: str, macrodict: Dict[str, str]) -> str:
    newexpr = expr
    thematch = macrocallre.search(newexpr)
    while thematch:
        macrocall = thematch.group(1)
        macroname = macrocall[1:-1]
        if macroname in macrodict:
            thismacrocallre = re.compile(macrocall)
            newexpr = thismacrocallre.sub(macrodict[macroname], newexpr)
            thematch = macrocallre.search(newexpr)
        else:
            settings.LOGGER.error('Unknown macro call encountered: {}.'.format(macroname))
            break
    return newexpr


macrodir = op.join(settings.SD_DIR, 'data', 'macros')
macrofilenames = [op.join(macrodir, 'sastamacros1.txt'), op.join(macrodir, 'sastamacros2.txt'), op.join(macrodir, 'newimperatives.txt')]

macrodict = generatemacros()
for macrofilename in macrofilenames:
    macrofile = open(macrofilename, 'r', encoding='utf8')
    macrodict = readmacros(macrofile, macrodict)
